#!/usr/bin/env python
"""splice: wrapper around the splice() syscall"""

from __future__ import print_function

from .utils import UnknownError
from cffi import FFI as _FFI
import errno as _errno

_ffi = _FFI()
_ffi.cdef("""
#define SPLICE_F_MOVE     ... /* This is a noop in modern kernels and is left here for compatibility */
#define SPLICE_F_NONBLOCK ... /* Make splice operations Non blocking (as long as the fd's are non blocking) */
#define SPLICE_F_MORE     ... /* After splice() more data will be sent, this is a hint to add TCP_CORK like buffering */
#define SPLICE_F_GIFT     ... /* unused for splice() (vmsplice compatibility) */

#define IOV_MAX ... /* Maximum ammount of vectors that can be written by vmsplice in one go */

struct iovec {
    void *iov_base; /* Starting address */
    size_t iov_len; /* Number of bytes */
};

ssize_t splice(int fd_in, signed long long *off_in, int fd_out, signed long long *off_out, size_t len, unsigned int flags);
ssize_t tee(int fd_in, int fd_out, size_t len, unsigned int flags);
ssize_t vmsplice(int fd, const struct iovec *iov, unsigned long nr_segs, unsigned int flags);

char * convert_str_to_void(char * buf);
""")

_C = _ffi.verify("""
#include <limits.h> /* used to define IOV_MAX */
#include <fcntl.h>
#include <sys/uio.h>

/* Its really hard in cffi to convert a python string to a char * WITHOUT using a function
   so instead lets just make a dummy function and use that. while we are at it, lets do
   the conversion to a void pointer as well so we dont need to much with types
*/
void * convert_str_to_void(char * buf){
    /* Take a string and convert it over to a void pointer */
    /* While simple, this is a work around for the cffi lib in python */
    return (void *)buf;
};
""", libraries=[], ext_package="butter")

def splice(fd_in, fd_out, in_offset=0, out_offset=0, len=0, flags=0):
    """Take data from fd_in and pass it to fd_out without going through userspace
    
    Arguments
    ----------
    :param file fd_in: File object or fd to splice from
    :param file fd_out: File object or fd to splice to
    :param int in_offset: Offset inside fd_in to read from
    :param int out_offset: Offset inside fd_out to write to
    :param int len: Ammount of data to transfer
    :param int flags: Flags to specify extra options
    
    Flags
    ------
    SPLICE_F_MOVE: This is a noop in modern kernels and is left here for compatibility
    SPLICE_F_NONBLOCK: Make splice operations Non blocking (as long as the fd's are non blocking)
    SPLICE_F_MORE: After splice() more data will be sent, this is a hint to add TCP_CORK like buffering
    SPLICE_F_GIFT: unused for splice() (vmsplice compatibility)
    
    Returns
    --------
    :return: Number of bytes written
    :rtype: int
    
    Exceptions
    -----------
    :raises ValueError: One of the file descriptors is unseekable
    :raises ValueError: Neither descriptor refers to a pipe
    :raises ValueError: Target filesystem does not support splicing
    :raises OSError: supplied fd does not refer to a file
    :raises OSError: Incorrect mode for file
    :raises MemoryError: Insufficient kernel memory
    :raises OSError: No writers waiting on fd_in
    :raises OSError: one or both fd's are in blocking mode and SPLICE_F_NONBLOCK specified
    """
    if hasattr(fd_in, 'fileno'):
        fd_in = fd_in.fileno()
    if hasattr(fd_out, 'fileno'):
        fd_out = fd_out.fileno()

    assert isinstance(fd_in, int), 'fd_in must be an integer'
    assert isinstance(fd_out, int), 'fd_in must be an integer'
    assert isinstance(in_offset, int), 'in_offset must be an integer'
    assert isinstance(out_offset, int), 'out_offset must be an integer'
    assert isinstance(len, int), 'len must be an integer'
    assert isinstance(flags, int), 'flags must be an integer'
    
    in_offset = _ffi.cast("long long *", in_offset)
    out_offset = _ffi.cast("long long *", out_offset)

    size = _C.splice(fd_in, in_offset, fd_out, out_offset, len, flags)
    
    if size < 0:
        err = _ffi.errno
        if err == _errno.EINVAL:
            if in_offset or out_offset:
                raise ValueError("fds may not be seekable")
            else:
                raise ValueError("Target filesystem does not support slicing or file may be in append mode")
        elif err == _errno.EBADF:
            raise ValueError("fds are invalid or incorrect mode for file")
        elif err == _errno.EPIPE:
            raise ValueError("offset specified but one of the fds is a pipe")
        elif err == _errno.ENOMEM:
            raise MemoryError("Insufficent kernel memory available")
        elif err == _errno.EAGAIN:
            raise OSError("No writers on fd_in or a fd is open in BLOCKING mode and NON_BLOCK specified to splice()")
        else:
            # If you are here, its a bug. send us the traceback
            raise UnknownError(err)

    return size


def tee(fd_in, fd_out, len=0, flags=0):
    """Splice data like the :py:func:`.splice` but also leave a copy of the data in the original fd's buffers
    
    Arguments
    ----------
    :param file fd_in: File object or fd to splice from
    :param file fd_out: File object or fd to splice to
    :param int len: Ammount of data to transfer
    :param int flags: Flags to specify extra options
    
    Flags
    ------
    SPLICE_F_MOVE: This is a noop in modern kernels and is left here for compatibility
    SPLICE_F_NONBLOCK: Make tee operations Non blocking (as long as the fd's are non blocking)
    SPLICE_F_MORE: unused for tee()
    SPLICE_F_GIFT: unused for tee() (:py:func:`.vmsplice` compatibility)
    
    Returns
    --------
    :return: Number of bytes written
    :rtype: int
    
    Exceptions
    -----------
    :raises ValueError: One of the file descriptors is not a pipe
    :raises ValueError: Both file descriptors refer to the same pipe
    :raises MemoryError: Insufficient kernel memory
    """
    if hasattr(fd_in, 'fileno'):
        fd_in = fd_in.fileno()
    if hasattr(fd_out, 'fileno'):
        fd_out = fd_out.fileno()
    
    assert isinstance(fd_in, int), 'fd_in must be an integer'
    assert isinstance(fd_out, int), 'fd_in must be an integer'
    assert isinstance(len, int), 'len must be an integer'
    assert isinstance(flags, int), 'flags must be an integer'

    size = _C.tee(fd_in, fd_out, len, flags)
    
    if size < 0:
        err = _ffi.errno
        if err == _errno.EINVAL:
            raise ValueError("fd_in or fd_out are not a pipe or refer to the same pipe")
        elif err == _errno.ENOMEM:
            raise MemoryError("Insufficent kernel memory available")
        else:
            # If you are here, its a bug. send us the traceback
            raise UnknownError(err)

    return size


def vmsplice(fd, vec, flags=0):
    """Write a list strings or byte buffers to the specified fd
    
    Arguments
    ----------
    :param file fd: File object or fd to write to
    :param list vec: A list of strings to write to the pipe
    :param int flags: Flags to specify extra options
    
    Flags
    ------
    SPLICE_F_MOVE: This is a noop in modern kernels and is left here for compatibility
    SPLICE_F_NONBLOCK: Make vmsplice operations Non blocking (as long as the fd is non blocking)
    SPLICE_F_MORE: unused for vmsplice()
    SPLICE_F_GIFT: Pass ownership of the pages to the kernel. You must not modify data in place
                   if using this option as the pages now belong to the kernel and bad things (tm)
                   will happen
                   if used, pages must be page aligned in both length and position
    Returns
    --------
    :return: Number of bytes written
    :rtype: int
    
    Exceptions
    -----------
    :raises ValueError: One of the file descriptors is not a pipe
    :raises ValueError: Both file descriptors refer to the same pipe
    :raises MemoryError: Insufficient kernel memory
    """
    if hasattr(fd, 'fileno'):
        fd = fd.fileno()

    assert isinstance(fd, int), 'fd must be an integer'
    assert isinstance(flags, int), 'flags must be an integer'

    n_vec =_ffi.new('struct iovec[]', len(vec))
    for str, v in zip(vec, n_vec) :
        v.iov_base = _C.convert_str_to_void(str)
        v.iov_len = len(str)
    
    size = _C.vmsplice(fd, n_vec, len(vec), flags)

    if size < 0:
        err = _ffi.errno
        if err == _errno.EBADF:
            raise ValueError("fd is not valid or does not refer to a pipe")
        if err == _errno.EINVAL:
            raise ValueError("nr_segs is 0 or greater than IOV_MAX; or memory not aligned if SPLICE_F_GIFT set")
        elif err == _errno.ENOMEM:
            raise MemoryError("Insufficent kernel memory available")
        else:
            # If you are here, its a bug. send us the traceback
            raise UnknownError(err)

    return size


SPLICE_F_MOVE = _C.SPLICE_F_MOVE    
SPLICE_F_NONBLOCK = _C.SPLICE_F_NONBLOCK
SPLICE_F_MORE = _C.SPLICE_F_MORE    
SPLICE_F_GIFT = _C.SPLICE_F_GIFT    

IOV_MAX = _C.IOV_MAX



#!/usr/bin/env python
"""inotify: Wrapper around the inotify syscalls providing both a function based and file like interface"""

from .utils import get_buffered_length as _get_buffered_length
from .utils import Eventlike as _Eventlike
from .utils import CLOEXEC_DEFAULT as _CLOEXEC_DEFAULT

from ._inotify import inotify_init, inotify_add_watch, inotify_rm_watch
from ._inotify import str_to_events
from ._inotify import event_name

from errno import EINVAL as _EINVAL

import os as _os

# Import all the constants
from ._inotify import C as _C
_l = locals()
for key in dir(_C):
    if key.startswith('IN_'):
        _l[key] = getattr(_C, key)
del key, _C, _l

class Inotify(_Eventlike):
    def __init__(self, flags=0, closefd=_CLOEXEC_DEFAULT):
        super(self.__class__, self).__init__()
        fd = inotify_init(flags, closefd=closefd)
        self._fd = fd
        
        self._events = []

        if flags & IN_NONBLOCK:
            self._blocking = False
        else:
            self._blocking = True
        
    def watch(self, path, events):
        wd = inotify_add_watch(self.fileno(), path, events)
        
        return wd
        
    def del_watch(self, wd):
        self.ignore(wd)

    def ignore(self, wd):
        inotify_rm_watch(self.fileno(), wd)
        
    def _read_events(self):
        fd = self.fileno()
        
        # The following code is complex but required to get the same values with
        # expected behavior with blocking/non-blocking fd's
        # blockers will block, non-blockers will return []
        buf_len = _get_buffered_length(fd)
        if buf_len == 0:
            try:
                _os.read(fd, 0)
            except OSError as err:
                # eat the OS error as the ammount of bytes in the buffer may have increased from 0
                # between the _get_buffered_length and here, we eat the exception and try again
                # so the caller does not ahve to deal with it. if there is a greater issue
                # the second read should blow up
                if err.errno != _EINVAL:
                    raise
                pass
            buf_len = _get_buffered_length(fd)
        raw_events = _os.read(fd, buf_len)

        events = str_to_events(raw_events)

        return events

def watch(path, events=IN_ALL_EVENTS):
    """Quick Convience function to watch a file or dir for any changes

    If a dir argument is provided this call will not recursively watch the directories
    due to limitations in inotify's API. if you wish to watch directories recursively
    you will need to recurs it yourself and add the watches manually

    Warning: if using this function to watch a file or dir repeatedly you may miss events
    due to a race condition, consider using the Inotify object instead to get all the 
    file events
    
    This function will not watch directories recursively if given a directory as an 
    argument and will only return events occurring in that directory
    
    Arguments
    ----------
    :param str path: the file/dir to watch for events to occur on
    :param int events: The inotify IN_* events to watch path for

    Returns
    --------
    :return: The file descriptor representing the signalfd
    :rtype: int

    """
    inotify = Inotify()
    try:
        wd = inotify.watch(path, events)
        event = inotify.wait()
    finally:
        inotify.close()
    
    return event



#!/usr/bin/env python
"""eventfd: maintain an atomic counter inside a file descriptor"""

from .utils import Eventlike as _Eventlike

from ._eventfd import EFD_CLOEXEC, EFD_NONBLOCK, EFD_SEMAPHORE
from ._eventfd import str_to_events, event_to_str
from ._eventfd import eventfd
from .utils import CLOEXEC_DEFAULT as _CLOEXEC_DEFAULT
import os as _os

class Eventfd(_Eventlike):
    def __init__(self, inital_value=0, flags=0, closefd=_CLOEXEC_DEFAULT):
        """Create a new Eventfd object

        Arguments
        ----------
        :param int inital_value: The inital value to set the eventfd to
        :param int flags: Flags to specify extra options
        
        Flags
        ------
        EFD_CLOEXEC: Close the eventfd when executing a new program
        EFD_NONBLOCK: Open the socket in non-blocking mode
        EFD_SEMAPHORE: Provide semaphore like semantics for read operations
        """
        super(self.__class__, self).__init__()
        self._fd = eventfd(inital_value, flags, closefd=closefd)
        
    def increment(self, value=1):
        """Increment the counter by the specified value
        
        :param int value: The value to increment the counter by (default=1)
        """
        assert value >= 0, "Value must be a positive number"
        
        packed_value = event_to_str(value)
        _os.write(self.fileno(), packed_value)

    def _read_events(self):
        """Read the current value of the counter and zero the counter

        Returns
        --------
        :return: The current count of the timer
        :rtype: int
        """
        data = _os.read(self.fileno(), 8)
        events = str_to_events(data)

        return events


    def __int__(self):
        return self.read_event()



#!/usr/bin/env python
"""eventfd: maintain an atomic counter inside a file descriptor"""
from .utils import UnknownError, CLOEXEC_DEFAULT
from cffi import FFI
import errno

ffi = FFI()
ffi.cdef("""
#define EFD_CLOEXEC ...
#define EFD_NONBLOCK ...
#define EFD_SEMAPHORE ...

int eventfd(unsigned int initval, int flags);
""")

C = ffi.verify("""
#include <sys/eventfd.h>
#include <stdint.h> /* Definition of uint64_t */
""", libraries=[], ext_package="butter")

EFD_CLOEXEC = C.EFD_CLOEXEC
EFD_NONBLOCK = C.EFD_NONBLOCK
EFD_SEMAPHORE = C.EFD_SEMAPHORE

def eventfd(inital_value=0, flags=0, closefd=CLOEXEC_DEFAULT):
    """Create a new eventfd
    
    Arguments
    ----------
    :param int inital_value: The inital value to set the eventfd to
    :param int flags: Flags to specify extra options
    :param bool closefd: Close the fd when a new process is exec'd
        
    Flags
    ------
    EFD_CLOEXEC: Close the eventfd when executing a new program
    EFD_NONBLOCK: Open the socket in non-blocking mode
    EFD_SEMAPHORE: Provide semaphore like semantics for read operations
    
    Returns
    --------
    :return: The file descriptor representing the eventfd
    :rtype: int
    
    Exceptions
    -----------
    :raises ValueError: Invalid value in flags
    :raises OSError: Max per process FD limit reached
    :raises OSError: Max system FD limit reached
    :raises OSError: Could not mount (internal) anonymous inode device
    :raises MemoryError: Insufficient kernel memory
    """
    assert isinstance(inital_value, int), "Inital value must be an integer"
    assert inital_value >= 0, "Inital value must be a positive number"
    assert isinstance(flags, int), "Flags must be an integer"
    
    if closefd:
        flags |= EFD_CLOEXEC
        
    fd = C.eventfd(inital_value, flags)
    
    if fd < 0:
        err = ffi.errno
        if err == errno.EINVAL:
            raise ValueError("Invalid value in flags")
        elif err == errno.EMFILE:
            raise OSError("Max per process FD limit reached")
        elif err == errno.ENFILE:
            raise OSError("Max system FD limit reached")
        elif err == errno.ENODEV:
            raise OSError("Could not mount (internal) anonymous inode device")
        elif err == errno.ENOMEM:
            raise MemoryError("Insufficent kernel memory available")
        else:
            # If you are here, its a bug. send us the traceback
            raise UnknownError(err)

    return fd

def str_to_events(str):
    value = ffi.new('uint64_t[1]')
    ffi.buffer(value, 8)[0:8] = str

    return [value[0]] # this may seem redundent but the original
                      # container is not actually a list

def event_to_str(event):
    # We use ffi rather than the array module as
    # python2.7 does not have an unsigned 64 bit in type
    event = ffi.new('uint64_t[1]', (event,))
    packed_event = ffi.buffer(event)[:]
    
    return packed_event



#!/usr/bin/env python

from select import select as _select
from cffi import FFI as _FFI
from os import close as _close
from collections import deque
import fcntl
import array
import errno

import platform

# Hack to backport PermissionError to older python versions
if platform.python_version_tuple() < ('3', '0', '0'):
    class PermissionError(OSError):
        """You do not have the required pemissions to use this syscall (CAP_SYS_ADMIN)"""
        pass
    class TimeoutError(Exception):
        """Timeout expired."""
        pass
    CLOEXEC_DEFAULT = True
else:
    CLOEXEC_DEFAULT = False

PermissionError = PermissionError
TimeoutError = TimeoutError

class InternalError(Exception):
    """This Error occured due to an internal bug or OS misconfiguration"""

class UnknownError(Exception):
    """This error is not handled and is unexpected by this library"""
    def __init__(self, errno):
        self.errno = errno
    
    def __str__(self):
        error_name = errno.errorcode.get(self.errno, "UNKNOWN")
        return "{}: Error: {} ({})".format(self.__doc__, self.errno, error_name)

_ffi = _FFI()
_ffi.cdef("""
#define FIONREAD ...
""")

_C = _ffi.verify("""
#include <sys/ioctl.h>
""", libraries=[], ext_package="butter")

def get_buffered_length(fd):
    buf = array.array("I", [0])
    fcntl.ioctl(fd, _C.FIONREAD, buf)
    return buf[0]
            

class Eventlike(object):
    _fd = None
    def __init__(self, *args, **kwargs):
        """*** This is a cooprative superclass, ensure you use super in the subclass's __init__ ***
        eg: super(self.__class__, self).__init__(*args, **kwargs)
        """
        self._events = []
        super(Eventlike, self).__init__()
    
    def close(self):
        _close(self.fileno())
        self._fd = None

    def fileno(self):
        if self._fd:
            return self._fd
        else:
            raise ValueError("I/O operation on closed file")

    def wait(self, timeout=None):
        if not self._events:
            # we use select here as the FD may be opened in non blocking mode
            rd, _, _ =_select([self], [], [], timeout)
            if self not in rd:
                raise TimeoutError("No event occured")

        return self.read_event()

    def closed(self):
        return False if self._fd else True

    def isatty(self):
        return False

    def mode(self):
        return getattr(self, '_mode', "r")

    def name(self):
        return repr(self)

    def read(self):
        raise NotImplementedError

    def readable(self):
        mode = getattr(self, '_mode', "r")
        return True if 'r' in mode or '+' in mode else False

    def readlines(self):
        raise NotImplementedError

    def seek(self):
        raise NotImplementedError

    def seekable(self):
        return False

    def tell(self):
        return 0

    def truncate(self):
        """Discard all events in the queue"""
        self._events = []

    def write(self):
        raise NotImplementedError

    def writable(self):
        mode = getattr(self, '_mode', "r")
        return True if 'w' in mode else False

    def writelines(self):
        raise NotImplementedError

    def __repr__(self):
        fd = "closed" if self.closed() else self.fileno()
        return "<{} fd={}>".format(self.__class__.__name__, fd)

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            if other._fd == self._fd:
                return True
        return False

    def __ne__(self, other):
        if isinstance(other, self.__class__):
            if other._fd != self._fd:
                return True
        return False

    def __hash__(self):
        return hash(self.__class__) ^ hash(self._fd)


    ### Event like behavior ###
    def __iter__(self):
        while True:
            yield self.wait()

    def read_event(self):
        """Return a single event, may read more than one event from the kernel and cache the values
        """
        try:
            event = self._events.pop(0)
        except IndexError:
            events = self._read_events()
            event = events.pop(0)
            self._events = events

        return event

    def read_events(self):
        """Read and return multiple events from the kernel
        """
        events = self._events
        self._events = []
        if len(events) > 0:
            return events
        else:
            return self._read_events()




#!/usr/bin/env python
"""fanotify: wrapper aroudn the fanotify family of syscalls for watching for file modifcation"""
from __future__ import print_function

from cffi import FFI as _FFI
from os import O_RDONLY, O_WRONLY, O_RDWR
from os import fdopen
import errno as _errno

from warnings import warn

warn(PendingDeprecationWarning(), "The seccomp module is deprecated in favour of the official bindings, please consider upgrading")

READ_EVENTS_MAX = 10

_ffi = _FFI()
_ffi.cdef("""

typedef void * scmp_filter_ctx;

scmp_filter_ctx seccomp_init(uint32_t def_action);
int seccomp_reset(scmp_filter_ctx ctx, uint32_t def_action);
void seccomp_release(scmp_filter_ctx ctx);
int seccomp_merge(scmp_filter_ctx dst, scmp_filter_ctx src);
int seccomp_load(scmp_filter_ctx ctx);

typedef uint64_t scmp_datum_t;

struct scmp_arg_cmp {
    unsigned int arg;
    enum scmp_compare op;
    scmp_datum_t datum_a;
    scmp_datum_t datum_b;
};


int seccomp_rule_add(scmp_filter_ctx ctx, uint32_t action,
                     int syscall, unsigned int arg_cnt, ...);
int seccomp_rule_add_exact(scmp_filter_ctx ctx, uint32_t action,
                           int syscall, unsigned int arg_cnt, ...);

int seccomp_rule_add_array(scmp_filter_ctx ctx,
                           uint32_t action, int syscall,
                           unsigned int arg_cnt,
                           const struct scmp_arg_cmp *arg_array);
int seccomp_rule_add_exact_array(scmp_filter_ctx ctx,
                                 uint32_t action, int syscall,
                                 unsigned int arg_cnt,
                                 const struct scmp_arg_cmp *arg_array);

//# Default actions
#define SCMP_ACT_KILL ...
#define SCMP_ACT_TRAP ...
#define SCMP_ACT_ALLOW ...
uint64_t _SCMP_ACT_ERRNO(uint64_t code);
uint64_t _SCMP_ACT_TRACE(uint64_t code);

//#Valid comparison op values are as follows:
#define SCMP_CMP_NE ...
#define SCMP_CMP_LT ...
#define SCMP_CMP_LE ...
#define SCMP_CMP_EQ ...
#define SCMP_CMP_GE ...
#define SCMP_CMP_GT ...
#define SCMP_CMP_MASKED_EQ ...

//#SCMP_SYS macro wont work how we want it, grab the syscall numbers manually
#define __NR_SCMP_ERROR ...
#define __NR_socket ...
#define __NR_bind ...
#define __NR_connect ...
#define __NR_listen ...
#define __NR_accept ...
#define __NR_getsockname ...
#define __NR_getpeername ...
#define __NR_socketpair ...
#define __NR_send ...
#define __NR_recv ...
#define __NR_sendto ...
#define __NR_recvfrom ...
#define __NR_shutdown ...
#define __NR_setsockopt ...
#define __NR_getsockopt ...
#define __NR_sendmsg ...
#define __NR_recvmsg ...
#define __NR_accept4 ...
#define __NR_recvmmsg ...
#define __NR_sendmmsg ...
#define __NR_semop ...
#define __NR_semget ...
#define __NR_semctl ...
#define __NR_semtimedop ...
#define __NR_msgsnd ...
#define __NR_msgrcv ...
#define __NR_msgget ...
#define __NR_msgctl ...
#define __NR_shmat ...
#define __NR_shmdt ...
#define __NR_shmget ...
#define __NR_shmctl ...
#define __NR_arch_prctl ...
#define __NR_bdflush ...
#define __NR_break ...
#define __NR_chown32 ...
#define __NR_epoll_ctl_old ...
#define __NR_epoll_wait_old ...
#define __NR_fadvise64_64 ...
#define __NR_fchown32 ...
#define __NR_fcntl64 ...
#define __NR_fstat64 ...
#define __NR_fstatat64 ...
#define __NR_fstatfs64 ...
#define __NR_ftime ...
#define __NR_ftruncate64 ...
#define __NR_getegid32 ...
#define __NR_geteuid32 ...
#define __NR_getgid32 ...
#define __NR_getgroups32 ...
#define __NR_getresgid32 ...
#define __NR_getresuid32 ...
#define __NR_getuid32 ...
#define __NR_gtty ...
#define __NR_idle ...
#define __NR_ipc ...
#define __NR_lchown32 ...
#define __NR__llseek ...
#define __NR_lock ...
#define __NR_lstat64 ...
#define __NR_mmap2 ...
#define __NR_mpx ...
#define __NR_newfstatat ...
#define __NR__newselect ...
#define __NR_nice ...
#define __NR_oldfstat ...
#define __NR_oldlstat ...
#define __NR_oldolduname ...
#define __NR_oldstat ...
#define __NR_olduname ...
#define __NR_prof ...
#define __NR_profil ...
#define __NR_readdir ...
#define __NR_security ...
#define __NR_sendfile64 ...
#define __NR_setfsgid32 ...
#define __NR_setfsuid32 ...
#define __NR_setgid32 ...
#define __NR_setgroups32 ...
#define __NR_setregid32 ...
#define __NR_setresgid32 ...
#define __NR_setresuid32 ...
#define __NR_setreuid32 ...
#define __NR_setuid32 ...
#define __NR_sgetmask ...
#define __NR_sigaction ...
#define __NR_signal ...
#define __NR_sigpending ...
#define __NR_sigprocmask ...
#define __NR_sigreturn ...
#define __NR_sigsuspend ...
#define __NR_socketcall ...
#define __NR_ssetmask ...
#define __NR_stat64 ...
#define __NR_statfs64 ...
#define __NR_stime ...
#define __NR_stty ...
#define __NR_truncate64 ...
#define __NR_tuxcall ...
#define __NR_ugetrlimit ...
#define __NR_ulimit ...
#define __NR_umount ...
#define __NR_vm86 ...
#define __NR_vm86old ...
#define __NR_waitpid ...
#define __NR_create_module ...
#define __NR_get_kernel_syms ...
#define __NR_get_thread_area ...
#define __NR_nfsservctl ...
#define __NR_query_module ...
#define __NR_set_thread_area ...
#define __NR__sysctl ...
#define __NR_uselib ...
#define __NR_vserver ...
#define __NR_arm_fadvise64_64 ...
#define __NR_arm_sync_file_range ...
#define __NR_finit_module ...
#define __NR_pciconfig_iobase ...
#define __NR_pciconfig_read ...
#define __NR_pciconfig_write ...
#define __NR_sync_file_range2 ...
#define __NR_syscall ...
#define __NR_afs_syscall ...
#define __NR_fadvise64 ...
#define __NR_getpmsg ...
#define __NR_ioperm ...
#define __NR_iopl ...
#define __NR_kcmp ...
#define __NR_migrate_pages ...
#define __NR_modify_ldt ...
#define __NR_putpmsg ...
#define __NR_sync_file_range ...
""")

_C = _ffi.verify("""
#include <seccomp.h>

uint64_t _SCMP_ACT_ERRNO(uint64_t code) {
    return 0x7ff00000U | (code & 0x0000ffffU);
}
uint64_t _SCMP_ACT_TRACE(uint64_t code) {
    return 0x00050000U | (code & 0x0000ffffU);
}

""", libraries=['seccomp'], ext_package="butter")


def condition(arg=0, comparison=_C.SCMP_CMP_EQ, arg1=_ffi.NULL, arg2=_ffi.NULL):
    """Initalise a new struct_arg_cmp object
    :parama arg int: Which argument to the syscall the test should be performed on
    :parama comparison: The type of comparison to make (SCMP_CMP_*)
    :parama arg1 int: argument 1 for the comaprison
    :parama arg2 int: argument 2 for the comparison (not always used)
    """
    rule = _ffi.new('struct scmp_arg_cmp *')
    rule.arg = arg
    rule.op = comparison
    rule.datum_a = _ffi.cast('uint64_t', arg1)
    rule.datum_b = _ffi.cast('uint64_t', arg2)

    return rule



#!/usr/bin/env python

from .utils import PermissionError, UnknownError
from cffi import FFI
import errno

ffi = FFI()
ffi.cdef("""

#define CLONE_FS      ...
#define CLONE_NEWNS   ...
#define CLONE_NEWUTS  ...
#define CLONE_NEWIPC  ...
#define CLONE_NEWUSER ...
#define CLONE_NEWPID  ...
#define CLONE_NEWNET  ...

//#long __clone(unsigned long flags, void *child_stack, ...);
long __clone(unsigned long flags, void *child_stack,
             void *ptid, void *ctid, void *regs);

int unshare(int flags);
int setns(int fd, int nstype);
""")
  

C = ffi.verify("""  
#include <sched.h>
#include <unistd.h>
#include <sys/types.h>
#include <unistd.h>

// man page
//long __clone(unsigned long flags, void *child_stack, ...);
long __clone(unsigned long flags, void *child_stack,
             void *ptid, void *ctid,
             void *regs);
""", libraries=[], ext_package="butter")

CLONE_ALL = C.CLONE_NEWIPC  | \
            C.CLONE_NEWNET  | \
            C.CLONE_NEWNS   | \
            C.CLONE_NEWUTS  | \
            C.CLONE_NEWPID  | \
            C.CLONE_NEWUSER

CLONE_NEWNS = C.CLONE_NEWNS
CLONE_NEWUTS = C.CLONE_NEWUTS
CLONE_NEWIPC = C.CLONE_NEWIPC
CLONE_NEWUSER = C.CLONE_NEWUSER
CLONE_NEWPID = C.CLONE_NEWPID
CLONE_NEWNET = C.CLONE_NEWNET


def unshare(flags):
    """Unshare the current namespace and create a new one

    Arguments
    ----------
    :param int flags: The flags controlling which namespaces to unshare

    Flags
    ------
    CLONE_NEWNS: Unshare the mount namespace causing mounts in this namespace
                 to not be visible to the parent namespace
    CLONE_NEWUTS: Unshare the system hostname allowing it to be changed independently
                  to the rest of the system
    CLONE_NEWIPC: Unshare the IPC namespace 
    CLONE_NEWUSER: Unshare the UID space allowing UIDs to be remapped to the parent
    CLONE_NEWPID: Unshare the PID space allowing remapping of PIDs relative to the parent
    CLONE_NEWNET: Unshare the network namespace, creating a separate set of network
                  interfaces/firewall rules

    Exceptions
    -----------
    :raises ValueError: Invalid value in flags
    """
    fd = C.unshare(flags)

    if fd < 0:
        err = ffi.errno
        if err == errno.EINVAL:
            raise ValueError("Invalid value in flags")
        elif err == errno.EPERM:
            raise PermissionError("Process in chroot or has incorrect permissions")
        elif err == errno.EUSERS:
            raise PermissionError("CLONE_NEWUSER specified but max user namespace nesting has been reached")
        elif err == errno.ENOMEM:
            raise MemoryError("Insufficent kernel memory available")
        else:
            # If you are here, its a bug. send us the traceback
            raise UnknownError(err)

    return fd


### prctl

def main():
    import os, errno, sys
    
#    ret = C.__clone(CLONE_NEWNET|CLONE_NEWUTS|CLONE_NEWIPC|CLONE_NEWNS, ffi.NULL)
#    ret = C.__clone(CLONE_NEWNET|CLONE_NEWUTS|CLONE_NEWIPC|CLONE_NEWNS, ffi.NULL, ffi.NULL, ffi.NULL, ffi.NULL)

    ret = C.unshare(CLONE_NEWNET|CLONE_NEWUTS|CLONE_NEWIPC|CLONE_NEWNS)
#    ret = C.unshare(CLONE_ALL)
    if ret >= 0:
#        with open("/proc/self/uid_map", "w") as f:
#            f.write("0 0 1\n")
        os.execl('/bin/bash', 'bash')
    else:
        print(ret, ffi.errno, errno.errorcode[ffi.errno])
        print("failed")
        sys.exit(1)

if __name__ == "__main__":
    main()



#!/usr/bin/env python
"""timerfd: recive timing events on a file descriptor"""

from .utils import Eventlike as _Eventlike
from .utils import CLOEXEC_DEFAULT as _CLOEXEC_DEFAULT
from ._timerfd import TimerVal, timerfd, timerfd_gettime, timerfd_settime
from ._timerfd import TFD_CLOEXEC, TFD_NONBLOCK, TFD_TIMER_ABSTIME
from ._timerfd import CLOCK_REALTIME, CLOCK_MONOTONIC
from ._timerfd import ffi as _ffi
import os as _os

class Timer(_Eventlike, TimerVal):
    """Timer is both an event like object providing the file-like/event-like interface as well
    as a TimerVal object to allow setting of the periodicity and offset of the timer in a single
    interface without creating a separate throw away TimerVal object
    
    To use, instance an instance as per normal:
    
    >>> t = Timer()
    
    Then manipulate the timer using the TimerVal like interface:
    
    >>> t.occuring.every(seconds=1).after(seconds=500000000)
    
    Finally, use the Timer object as you would normally for an event like object:
    
    >>> for i, event in enumerate(t):
    ...     print ("Timer Fired")
    ...     if i >= 2: break
    
    This should print 'Timer Fired' after 0.5s followed by 1s after that

    read()ing and wait()ing on this socket returns an integer representing the amount
    of timer expiration's since the last time the timer was read. This can be used to make
    a high precision low overhead timer by using a repeating event with an expiry
    of 1ns and then using read() to get the amount of elapsed nano seconds. This has
    near 0% cpu overhead. Using the timer in this manner is refered to as an 'interval
    timer'
    """
    def __init__(self, clock_type=CLOCK_MONOTONIC, flags=0, closefd=_CLOEXEC_DEFAULT):
        """Create a new Timerfd object

        Arguments
        ----------
        :param int clock_type: The type of clock to use for timing
        :param int flags: Flags to specify extra options
        
        Flags
        ------
        CLOCK_REALTIME: Use a time source that matches the wall time
        CLOCK_MONOTONIC: Use a monotonically incrementing time source
        TFD_CLOEXEC: Close the timerfd when executing a new program
        TFD_NONBLOCK: Open the socket in non-blocking mode
        """
        super(Timer, self).__init__()
        self._fd = timerfd(clock_type, flags, closefd=closefd)
    
    def get_current(self):
        """Retrives the current values of the timer from the kernel
        
        Returns
        --------
        :return: The old timer value
        :rtype: TimerVal
        """
        return timerfd_gettime(self.fileno())

    def update(self, absolute=False):
        """Update the kernel with the current values for the timer
        
        Arguments
        ---------
        :param bool absolute: Determines if the values in the timer should be considered absolute
        (seconds since UNIX epoch) or if they should be added to the current time to determine
        when the next event occurs
        
        Returns
        --------
        :return: The old timer value
        :rtype: TimerVal
        """
        flags = TFD_TIMER_ABSTIME if absolute else 0
        old_timer = timerfd_settime(self.fileno(), self._timerspec, flags)
        
        return old_timer
    
    def _read_events(self):
        data = _os.read(self.fileno(), 8)
        value = _ffi.new('uint64_t[1]')
        _ffi.buffer(value, 8)[0:8] = data

        return [value[0]] # value's container is not a list
                          # lets make it one to expose a fammliar
                          # interface
    def __repr__(self):
            fd = "closed" if self.closed() else self.fileno()
            return "<{} fd={} offset=({}s, {}ns) reoccuring=({}s, {}ns)>".format(self.__class__.__name__,
                                                                           fd,
                                                                           self._timerspec.it_value.tv_sec,
                                                                           self._timerspec.it_value.tv_nsec,
                                                                           self._timerspec.it_interval.tv_sec,
                                                                           self._timerspec.it_interval.tv_nsec)




#!/usr/bin/env python
"""fanotify: wrapper around the fanotify family of syscalls for watching for file modifcation"""

from .utils import get_buffered_length as _get_buffered_length
from .utils import Eventlike as _Eventlike
from .utils import CLOEXEC_DEFAULT as _CLOEXEC_DEFAULT

from os import O_RDONLY, O_WRONLY, O_RDWR
from os import read as _read

from ._fanotify import fanotify_init, fanotify_mark, str_to_events

# Import all the constants
from ._fanotify import C as _C
_l = locals()
for key in dir(_C):
    if key.startswith('FAN_'):
        _l[key] = getattr(_C, key)
del key, _C, _l

class Fanotify(_Eventlike):
    blocking = True
    
    def __init__(self, flags, event_flags=O_RDONLY, closefd=_CLOEXEC_DEFAULT):
        super(self.__class__, self).__init__()
        self._fd = fanotify_init(flags, event_flags, closefd=closefd)

        self._events = []

        if flags & FAN_NONBLOCK:
            self.blocking = false
        
        if event_flags & O_RDWR|O_WRONLY:
            self._mode = 'w+'
        elif event_flags & O_WRONLY:
            self._mode = 'w'
        else:
            self._mode = 'r'

    def watch(self, path, mask, flags=0, dfd=0):
        flags |= FAN_MARK_ADD
        fanotify_mark(self.fileno(), path, mask, flags, dfd)

    def del_watch(self, path, mask, flags=0, dfd=0):
        self.ignore(path, mask, flags, dfd)
        
    def ignore(self, path, mask, flags=0, dfd=0):
        flags |= FAN_MARK_REMOVE
        fanotify_mark(self.fileno(), path, mask, flags, dfd)

    def _read_events(self):
        fd = self.fileno()

        buf_len = _get_buffered_length(fd)
        raw_events = _read(fd, buf_len)

        events = str_to_events(raw_events)

        return events



#!/usr/bih/env python
from ..eventfd import Eventfd as _Eventfd
from collections import deque as _deque
import asyncio as _asyncio


class Eventfd_async:
    def __init__(self, inital_value=0, flags=0, *, loop=None):
        self._loop = loop or _asyncio.get_event_loop()
        self._eventfd = _Eventfd(inital_value, flags)
        self._getters = _deque()
        self._value = inital_value
        
    def increment(self, value=1):
        """Increment the counter by the specified value

        :param int value: The value to increment the counter by (default=1)
        """
        self._eventfd.increment(value)

    @_asyncio.coroutine
    def wait(self):
        """Read the current value of the counter and zero the counter

        Returns
        --------
        :return: The current count of the timer
        :rtype: int
        """
        self._loop.add_reader(self._eventfd.fileno(), self._read_event)

        waiter = _asyncio.Future(loop=self._loop)

        self._getters.append(waiter)

        return (yield from waiter)

    def get_last(self):
        return self._value

    def _consume_done_getters(self):
        # Delete waiters at the head of the get() queue who've timed out.
        while self._getters and self._getters[0].done():
            self._getters.popleft()

    def _read_event(self):
        value = self._eventfd.read_event()
        self._put_event(value)
    
    def _put_event(self, value):
        """Put an item into the queue without blocking.

        If no free slot is immediately available, raise QueueFull.
        """
        self._loop.remove_reader(self._eventfd.fileno())

        self._consume_done_getters()

        for getter in self._getters:
            getter.set_result(value)

        self._value = value

    def close(self):
        self._eventfd.close()

    def __repr__(self):
        fd = self._eventfd._fd or 'closed'
        return "<{} fd={} value={}>".format(self.__class__.__name__, fd, self._value)

def _watcher(loop):
    ev = Eventfd_async(10)
    print(ev)
    print('Waiting read:', (yield from ev.wait()))
    ev.increment(2)
    ev.increment(2)
    print('Waiting read:', (yield from ev.wait()))
    print('No wait read:', ev.get_last())
    ev.close()
    print(ev)
    print('done')

def _main():
    import logging
    import asyncio
    
    log = logging.getLogger()
    log.setLevel(logging.DEBUG)
    log.addHandler(logging.StreamHandler())
    
    loop = asyncio.get_event_loop()
    
    from asyncio import Task
    task = Task(_watcher(loop))
    
    loop.run_until_complete(task)
    

if __name__ == "__main__":
    _main()



#!/usr/bih/env python
from ..timerfd import CLOCK_REALTIME as _CLOCK_REALTIME
from ..timerfd import Timer as _Timer
from collections import deque as _deque
import asyncio as _asyncio

class Timerfd_async:
    def __init__(self, clock_type=_CLOCK_REALTIME, flags=0, *, loop=None):
        self._loop = loop or _asyncio.get_event_loop()
        self._timerfd = _Timer(clock_type, flags)
        self._getters = _deque()
        
        self.set_one_off = self._timerfd.set_one_off
        self.set_reoccuring = self._timerfd.set_reoccuring
        self.enabled = self._timerfd.__class__.__dict__['enabled']
        self.disabled = self._timerfd.__class__.__dict__['disabled']
        self.disable = self._timerfd.disable
        self.get_current = self._timerfd.get_current
        
    @_asyncio.coroutine
    def wait(self):
        """Wait for the timer to expire, returning how many events have 
        elappsed since the last call to wait()

        Returns
        --------
        :return: The current count of the timer
        :rtype: int
        """
        self._loop.add_reader(self._timerfd.fileno(), self._read_event)

        waiter = _asyncio.Future(loop=self._loop)

        self._getters.append(waiter)

        return (yield from waiter)

    def _consume_done_getters(self):
        # Delete waiters at the head of the get() queue who've timed out.
        while self._getters and self._getters[0].done():
            self._getters.popleft()

    def _read_event(self):
        value = self._timerfd.read_event()
        self._put_event(value)
    
    def _put_event(self, value):
        """Put an item into the queue without blocking.

        If no free slot is immediately available, raise QueueFull.
        """
        self._loop.remove_reader(self._timerfd.fileno())

        self._consume_done_getters()

        for getter in self._getters:
            getter.set_result(value)

    def close(self):
        self._timerfd.close()

    def __repr__(self):
        fd = self._timerfd._fd or "closed"
        return "<{} fd={}>".format(self.__class__.__name__, fd)

def watcher(loop):
    from asyncio import sleep
    from time import time
    
    t = Timerfd_async()
    print(t)
    
    time_val = 0.5
    t.set_reoccuring(time_val)
    print("Setting time interval to {:.2f} seconds".format(time_val))

    for i in range(5):
        old_time = time()
        num_events = yield from t.wait()
        new_time = time()
        assert num_events == 1, "Too many events"
        print("Woke up after {:.2f} seconds".format(new_time - old_time))

    print("Got all 5 events")

    print("Sleeping for 0.3s")
    yield from sleep(0.3)
    print("Next event:", t.get_current())
    
    t.close()
    print(t)


def main():
    import logging
    import asyncio
    
    log = logging.getLogger()
    log.setLevel(logging.DEBUG)
    log.addHandler(logging.StreamHandler())
    
    loop = asyncio.get_event_loop()
    
    from asyncio import Task
    task = Task(watcher(loop))
    
    loop.run_until_complete(task)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
"""inotify: Wrapper around the inotify syscalls providing both a function based and file like interface"""

from collections import namedtuple
from .utils import PermissionError, UnknownError, CLOEXEC_DEFAULT
from cffi import FFI
import errno


ffi = FFI()
ffi.cdef("""
/*
 * struct inotify_event - structure read from the inotify device for each event
 *
 * When you are watching a directory, you will receive the filename for events
 * such as IN_CREATE, IN_DELETE, IN_OPEN, IN_CLOSE, ..., relative to the wd.
 */
struct inotify_event {
        int           wd;
        uint32_t      mask;
        uint32_t      cookie;
        uint32_t      len;
//        char          name[0]; # we calculate this manually in str_to_event
};

/* the following are legal, implemented events that user-space can watch for */
#define IN_ACCESS        ...  /* File was accessed */
#define IN_MODIFY        ...  /* File was modified */
#define IN_ATTRIB        ...  /* Metadata changed */
#define IN_CLOSE_WRITE   ...  /* Writtable file was closed */
#define IN_CLOSE_NOWRITE ...  /* Unwrittable file closed */
#define IN_OPEN          ...  /* File was opened */
#define IN_MOVED_FROM    ...  /* File was moved from X */
#define IN_MOVED_TO      ...  /* File was moved to Y */
#define IN_CREATE        ...  /* Subfile was created */
#define IN_DELETE        ...  /* Subfile was deleted */
#define IN_DELETE_SELF   ...  /* Self was deleted */
#define IN_MOVE_SELF     ...  /* Self was moved */

/* the following are legal events.  they are sent as needed to any watch */
#define IN_UNMOUNT       ...  /* Backing fs was unmounted */
#define IN_Q_OVERFLOW    ...  /* Event queued overflowed */
#define IN_IGNORED       ...  /* File was ignored */

/* helper events */
#define IN_CLOSE         ...  /* close */
#define IN_MOVE          ...  /* moves */

/* special flags */
#define IN_ONLYDIR       ...  /* only watch the path if it is a directory */
#define IN_DONT_FOLLOW   ...  /* don't follow a sym link */
#define IN_EXCL_UNLINK   ...  /* exclude events on unlinked objects */
#define IN_MASK_ADD      ...  /* add to the mask of an already existing watch */
#define IN_ISDIR         ...  /* event occurred against dir */
#define IN_ONESHOT       ...  /* only send event once */

/*
 * All of the events - we build the list by hand so that we can add flags in
 * the future and not break backward compatibility.  Apps will get only the
 * events that they originally wanted.  Be sure to add new events here!
 */
#define IN_ALL_EVENTS  ...

/* Flags for sys_inotify_init1.  */
#define IN_CLOEXEC  ...
#define IN_NONBLOCK ...

int inotify_init(void);
int inotify_init1(int flags);
int inotify_add_watch(int fd, const char *pathname, uint32_t mask);
int inotify_rm_watch(int fd, int wd);
""")

C = ffi.verify("""
#include <sys/inotify.h>
#include <sys/ioctl.h>
""", libraries=[], ext_package="butter")


def inotify_init(flags=0, closefd=CLOEXEC_DEFAULT):
    """Initialise an inotify instnace and return a File Descriptor to refrence is
    
    Arguments:
    -----------
    Flags:
    -------
    IN_CLOEXEC: Automatically close the inotify handle on exec()
    IN_NONBLOCK: Place the file descriptor in non blocking mode
    """
    assert isinstance(flags, int), 'Flags must be an integer'

    if closefd:
        flags |= IN_CLOEXEC

    fd = C.inotify_init1(flags)
    
    if fd < 0:
        err = ffi.errno
        if err == errno.EINVAL:
            raise ValueError("Invalid argument or flag")
        elif err == errno.EMFILE:
            raise OSError("Maximum inotify instances reached")
        elif err == errno.ENFILE:
            raise OSError("File descriptor limit hit")
        elif err == errno.ENOMEM:
            raise MemoryError("Insufficent kernel memory avalible")
        else:
            # If you are here, its a bug. send us the traceback
            raise UnknownError(err)

    return fd
    
def inotify_add_watch(fd, path, mask):
    """Start watching a filepath for events
    
    Arguments:
    -----------
    fd:    The inotify file descriptor to attach the watch to
    path:  The path to the file/directory to be monitored for events
    mask:  The events to listen for
    
    Flags:
    -------
    IN_ACCESS:        File was accessed
    IN_MODIFY:        File was modified
    IN_ATTRIB:        Metadata changed
    IN_CLOSE_WRITE:   Writtable file was closed
    IN_CLOSE_NOWRITE: Unwrittable file closed
    IN_OPEN:          File was opened
    IN_MOVED_FROM:    File was moved from X
    IN_MOVED_TO:      File was moved to Y
    IN_CREATE:        Subfile was created
    IN_DELETE:        Subfile was deleted
    IN_DELETE_SELF:   Self was deleted
    IN_MOVE_SELF:     Self was moved

    IN_ONLYDIR:      only watch the path if it is a directory
    IN_DONT_FOLLOW:  don't follow a sym link
    IN_EXCL_UNLINK:  exclude events on unlinked objects
    IN_MASK_ADD:     add to the mask of an already existing watch
    IN_ISDIR:        event occurred against dir
    IN_ONESHOT:      only send event once
    
    Returns:
    ---------
    int: A watch descriptor that can be passed to inotify_rm_watch
    
    Exceptions:
    ------------
    ValueError:
    * No valid events in the event mask
    * fd is not an inotify file descriptor
    OSError:
    * fd is not a valid file descriptor
    * Process has no access to specified file
    * File/Folder specified does not exist
    * Maximum number of watches hit
    MemoryError:
    * Raised if the kernel cannot allocate sufficent resources to handle the watch (eg kernel memory)
    """
    if hasattr(fd, "fileno"):
        fd = fd.fileno()

    assert isinstance(fd, int), "fd must by an integer"
    assert isinstance(path, (str, bytes)), "path is not a string"
    assert len(path) > 0, "Path must be longer than 0 chars"
    assert isinstance(mask, int), "mask must be an integer"
    
    if isinstance(path, str):
        path = path.encode()
    
    wd = C.inotify_add_watch(fd, path, mask)

    if wd < 0:
        err = ffi.errno
        if err == errno.EINVAL:
            raise ValueError("The event mask contains no valid events; or fd is not an inotify file descriptor")
        elif err == errno.EACCES:
            raise PermissionError("You do not have permission to read the specified path")
        elif err == errno.EBADF:
            raise ValueError("fd is not a valid file descriptor")
        elif err == errno.EFAULT:
            raise ValueError("path points to a file/folder outside the processes accessible address space")
        elif err == errno.ENOENT:
            raise ValueError("File/Folder pointed to by path does not exist")
        elif err == errno.ENOSPC:
            raise OSError("Maximum number of watches hit or insufficent kernel resources")
        elif err == errno.ENOMEM:
            raise MemoryError("Insufficent kernel memory avalible")
        else:
            # If you are here, its a bug. send us the traceback
            raise UnknownError(err)
            
    return wd
    
def inotify_rm_watch(fd, wd):
    """Stop watching a path for events
    
    Arguments:
    -----------
    fd: The inotify file descriptor to remove the watch from
    wd: The Watch to be removed
    
    Returns:
    ---------
    None
    
    Exceptions:
    ------------
    ValueError: Returned if supplied watch is not valid or if the file descriptor is not an inotify file descriptor
    OSError: File descriptor is invalid
    """
    if hasattr(fd, 'fileno'):
        fd = fd.fileno()
            
    assert isinstance(fd, int), "fd must by an integer"
    assert isinstance(wd, int), "wd must be an integer"

    ret = C.inotify_rm_watch(fd, wd)

    if ret < 0:
        err = ffi.errno
        if err == errno.EINVAL:
            raise ValueError("wd is invalid or fd is not an inotify File Descriptor")
        elif err == errno.EBADF:
            raise ValueError("fd is not a valid file descriptor")
        else:
            # If you are here, its a bug. send us the traceback
            raise UnknownError(err)


def str_to_events(str):
    event_struct_size = ffi.sizeof('struct inotify_event')

    events = []

    str_buf = ffi.new('char[]', len(str))
    str_buf[0:len(str)] = str

    i = 0
    while i < len(str_buf):
        event = ffi.cast('struct inotify_event *', str_buf[i:i+event_struct_size])

        filename_start = i + event_struct_size
        filename_end = filename_start + event.len
        filename = ffi.string(str_buf[filename_start:filename_end])
        
        events.append(InotifyEvent(event.wd, event.mask, event.cookie, filename))
        
        i += event_struct_size + event.len

    return events


InotifyEvent = namedtuple("InotifyEvent", "wd mask cookie filename")
class InotifyEvent(InotifyEvent):
    __slots__ = []
    @property
    def access_event(self):
        return True if self.mask & IN_ACCESS else False

    @property
    def modify_event(self):
        return True if self.mask & IN_MODIFY else False

    @property
    def attrib_event(self):
        return True if self.mask & IN_ATTRIB else False

    @property
    def close_write_event(self):
        return True if self.mask & IN_CLOSE_WRITE else False

    @property
    def close_nowrite_event(self):
        return True if self.mask & IN_CLOSE_NOWRITE else False

    @property
    def close_event(self):
        return True if self.mask & (IN_CLOSE_NOWRITE|IN_CLOSE_WRITE) else False

    @property
    def open_event(self):
        return True if self.mask & IN_OPEN else False

    @property
    def moved_from_event(self):
        return True if self.mask & IN_MOVED_FROM else False

    @property
    def moved_to_event(self):
        return True if self.mask & IN_MOVED_TO else False

    @property
    def create_event(self):
        return True if self.mask & IN_CREATE else False

    @property
    def delete_event(self):
        return True if self.mask & IN_DELETE else False

    @property
    def delete_self_event(self):
        return True if self.mask & IN_DELETE_SELF else False

    @property
    def move_self_event(self):
        return True if self.mask & IN_MOVE_SELF else False

    @property
    def is_dir_event(self):
        return True if self.mask & IN_ISDIR else False

# update the local namespace with flags and provide
# a handy dict for reversable lookups
event_name = {}
_l = locals()
for key in dir(C):
    if key.startswith('IN_'):
        val = getattr(C, key)
        _l[key] = val
        event_name[key] = val
        event_name[val] = key



#!/usr/bin/env python
"""timerfd: recive timing events on a file descriptor

For new code use TimerVal instead of TimerSpec as TimerSpec is slated to be removed 
after the 1.0 release and provides a simplier and faster interface that is easier
to interpret
"""

from .utils import UnknownError, InternalError, CLOEXEC_DEFAULT
from collections import namedtuple
from cffi import FFI
import errno
import math

TimePair = namedtuple('TimePair', 'seconds nano_seconds')

ffi = FFI()
ffi.cdef("""
#define TFD_CLOEXEC ...
#define TFD_NONBLOCK ...

#define TFD_TIMER_ABSTIME ...

#define CLOCK_REALTIME ...
#define CLOCK_MONOTONIC ...

typedef long int time_t;

struct timespec {
    time_t tv_sec; /* Seconds */
    long tv_nsec; /* Nanoseconds */
};

struct itimerspec {
    struct timespec it_interval; /* Interval for periodic timer */
    struct timespec it_value; /* Initial expiration */
};

int timerfd_create(int clockid, int flags);

int timerfd_settime(int fd, int flags,
                    const struct itimerspec *new_value,
                    struct itimerspec *old_value);

int timerfd_gettime(int fd, struct itimerspec *curr_value);
""")

C = ffi.verify("""
#include <sys/timerfd.h>
#include <stdint.h> /* Definition of uint64_t */
#include <time.h>
""", libraries=[], ext_package="butter")


TFD_CLOEXEC = C.TFD_CLOEXEC
TFD_NONBLOCK = C.TFD_NONBLOCK
TFD_TIMER_ABSTIME = C.TFD_TIMER_ABSTIME

CLOCK_REALTIME = C.CLOCK_REALTIME
CLOCK_MONOTONIC = C.CLOCK_MONOTONIC

class TimerVal(object): 
    """ timer = TimerVal()
    timer.occuring.every(mins=5, nano_seconds=3).after(seconds=5)
    timer.offset(seconds=3).and_repeats.every(seconds=5)
    """
    __slots__ = ['_timerspec']
    def __init__(self, timerspec=None):
        if timerspec:
            self._timerspec = timerspec
        else:
            self._timerspec = ffi.new('struct itimerspec *')
        
        super(TimerVal, self).__init__()

    @property
    def occuring(self):
        return self
    @property
    def and_repeats(self):
        return self
    
    def every(self, seconds=None, nano_seconds=None):
        if seconds is not None:
            self._timerspec.it_interval.tv_sec = seconds
            # arm the timer
            if not self._timerspec.it_value.tv_sec:
                self._timerspec.it_value.tv_sec = seconds
        if nano_seconds is not None:
            self._timerspec.it_interval.tv_nsec = nano_seconds
            # arm the timer
            if not self._timerspec.it_value.tv_nsec:
                self._timerspec.it_value.tv_nsec = nano_seconds
        return self
        
    def repeat(self, seconds=None, nano_seconds=None):
        return self.every(seconds, nano_seconds)
        
    def repeats(self, seconds=None, nano_seconds=None):
        return self.every(seconds, nano_seconds)
    
    def after(self, seconds=None, nano_seconds=None):
        if seconds is not None:
            self._timerspec.it_value.tv_sec = seconds
        if nano_seconds is not None:
            self._timerspec.it_value.tv_nsec = nano_seconds
        return self
        
    def offset(self, seconds=None, nano_seconds=None):
        return self.after(seconds, nano_seconds)

    def __timerspec__(self):
        return self._timerspec

    def __bool__(self):
        return bool(self.it_value.tv_sec or self.it_value.tv_nsec)
    
    @property
    def enabled(self):
        """Will this timer fire if used?, returns a bool"""
        return bool(self)
    
    def disable(self):
        """Disable the timer from firing"""
        return self.offset(0, 0)

    @property
    def next_event(self):
        """returns a tuple containg (seconds, nanoseconds) representing when the timer next expires"""
        return TimePair(self._timerspec.it_value.tv_sec, self._timerspec.it_value.tv_nsec)

    @property
    def period(self):
        """returns a tuple containg (seconds, nanoseconds) representing how freqently the timer expires"""
        return TimePair(self._timerspec.it_interval.tv_sec, self._timerspec.it_interval.tv_nsec)
    
    def __repr__(self):
        return "<{} offset=({}s, {}ns) reoccuring=({}s, {}ns)>".format(self.__class__.__name__,
                                                                       self._timerspec.it_value.tv_sec, 
                                                                       self._timerspec.it_value.tv_nsec,
                                                                       self._timerspec.it_interval.tv_sec, 
                                                                       self._timerspec.it_interval.tv_nsec)


def timerfd(clock_type=CLOCK_MONOTONIC, flags=0, closefd=CLOEXEC_DEFAULT):
    """Create a new timerfd
    
    Arguments
    ----------
    :param int clock_type: The type of clock to use internally
                           DEFAULT: CLOCK_MONOTONIC
    :param int flags: Flags to specify extra options
    
    Flags
    ------
    CLOCK_REALTIME: Use a clock that mirrors the system time
                    (will be affected by settime)
    CLOCK_MONOTONIC: Use a Monotonically increasing clock value
    TFD_CLOEXEC: Close the timerfd when executing a new program
    TFD_NONBLOCK: Open the socket in non-blocking mode

    in nearly all simple cases you will want CLOCK_MONOTONIC. When
    using CLOCK_REALTIME you will need to be aware of and adjust for
    the clock moving back and forwards (eg ntp, admins, daylight savings,
    leap seconds, clock drift, time dilation as you approach the speed
    of light)
    
    Returns
    --------
    :return: The file descriptor representing the timerfd
    :rtype: int
    
    Exceptions
    -----------
    :raises ValueError: Invalid value in flags
    :raises OSError: Max per process FD limit reached
    :raises OSError: Max system FD limit reached
    :raises OSError: Could not mount (internal) anonymous inode device
    :raises MemoryError: Insufficient kernel memory
    """
    assert isinstance(clock_type, int), 'Clock Type must be an integer'
    assert isinstance(flags, int), 'Flags must be an integer'

    if closefd:
        flags |= TFD_CLOEXEC
        
    fd = C.timerfd_create(clock_type, flags)
    
    if fd < 0:
        err = ffi.errno
        if err == errno.EINVAL:
            if not (clock_type & CLOCK_MONOTONIC) and not (clock_type & CLOCK_REALTIME):
                raise ValueError("clock_type is not one of CLOCK_MONOTONIC or CLOCK_REALTIME")
            raise ValueError("Invalid value in flags")
        elif err == errno.EMFILE:
            raise OSError("Max per process FD limit reached")
        elif err == errno.ENFILE:
            raise OSError("Max system FD limit reached")
        elif err == errno.ENODEV:
            raise OSError("Could not mount (internal) anonymous inode device")
        elif err == errno.ENOMEM:
            raise MemoryError("Insufficent kernel memory available")
        else:
            # If you are here, its a bug. send us the traceback
            raise UnknownError(err)

    return fd


def timerfd_gettime(fd):
    """Get the current expiry time of a timerfd
    
    Arguments
    ----------
    :param int fd: File descriptor representing the timerfd

    Returns
    --------
    :return: The file descriptor representing the timerfd
    :rtype: int
    
    Exceptions
    -----------
    :raises ValueError: Invalid value in flags
    :raises OSError: Max per process FD limit reached
    :raises OSError: Max system FD limit reached
    :raises OSError: Could not mount (internal) anonymous inode device
    :raises MemoryError: Insufficient kernel memory
    """
    if hasattr(fd, 'fileno'):
        fd = fd.fileno()

    assert isinstance(fd, int), 'fd must be an integer'
    
    curr_val = ffi.new('struct itimerspec *')
    ret = C.timerfd_gettime(fd, curr_val)
    
    if ret < 0:
        err = ffi.errno
        if err == errno.EBADF:
            raise ValueError("fd is not a valid file descriptor")
        elif err == errno.EFAULT:
            raise InternalError("curr_val is not a valid pointer (internal/bug, let us know)")
        elif err == errno.EINVAL:
            raise ValueError("fd is not a valid timerfd")
        else:
            # If you are here, its a bug. send us the traceback
            raise UnknownError(err)

    curr_val = TimerVal(timerspec=curr_val)
    return curr_val


def timerfd_settime(fd, timer_spec, flags=0):
    """Set the expiry time of a timerfd
    
    Arguments
    ----------
    :param int fd: File descriptor representing the timerfd
    :param int inital_value: The inital value to set the timerfd to
    :param int flags: Flags to specify extra options
    
    Flags
    ------
    TFD_TIMER_ABSTIMER: The specified time is an absolute value rather than relative to now
        
    Returns
    --------
    :return: The file descriptor representing the timerfd
    :rtype: int
    
    Exceptions
    -----------
    :raises ValueError: Invalid value in flags
    :raises OSError: Max per process FD limit reached
    :raises OSError: Max system FD limit reached
    :raises OSError: Could not mount (internal) anonymous inode device
    :raises MemoryError: Insufficient kernel memory
    """
    if hasattr(fd, 'fileno'):
        fd = fd.fileno()

    assert isinstance(fd, int), 'fd must be an integer'
    assert isinstance(flags, int), 'flags must be an integer'
    
    if hasattr(timer_spec, '__timerspec__'):
        timer_spec = timer_spec.__timerspec__()

    assert isinstance(timer_spec, FFI.CData) # ensure passed in value is what we want
    
    old_timer_spec = ffi.new('struct itimerspec *')

    ret = C.timerfd_settime(fd, flags, timer_spec, old_timer_spec)
    
    if ret < 0:
        err = ffi.errno
        if err == errno.EINVAL:
            if timer_spec.it_interval.tv_sec > 999999999:
                raise ValueError("Repeat Seconds > 999,999,999")
            elif timer_spec.it_interval.tv_nsec > 999999999:
                raise ValueError("Repeat Nano seconds > 999,999,999")
            elif timer_spec.it_value.tv_sec > 999999999:
                raise ValueError("Offset Seconds > 999,999,999")
            elif timer_spec.it_value.tv_nsec > 999999999:
                raise ValueError("Offset Nano seconds > 999,999,999")
            else:
                raise ValueError('flags is invalid or fd not a timerfd')
        elif err == errno.EFAULT:
            raise InternalError("timer_spec does not point to a valid timer specfication")
        elif err == errno.EMFILE:
            raise OSError("Max per process FD limit reached")
        elif err == errno.ENFILE:
            raise OSError("Max system FD limit reached")
        elif err == errno.ENODEV:
            raise OSError("Could not mount (internal) anonymous inode device")
        elif err == errno.ENOMEM:
            raise MemoryError("Insufficent kernel memory available")
        else:
            # If you are here, its a bug. send us the traceback
            raise UnknownError(err)
            
    old_timer_spec = TimerVal(timerspec=old_timer_spec)
    return old_timer_spec



#!/usr/bin/env python
"""Butter: library to give python access to linux's more lower level features"""

__author__ = "Da_Blitz"
__version__ = "0.11"
__email__ = "code@pocketnix.org"
__license__ = "BSD (3 Clause)"
__url__ = "http://code.pocketnix.org/butter"

__all__ = ['fanotify', 'inotify', 'seccomp', 'splice', 'system', 'utils']



#!/usr/bin/env python
"""signalfd: Recive signals over a file descriptor"""

from .utils import UnknownError, CLOEXEC_DEFAULT
from cffi import FFI
import platform
import signal
import errno

ffi = FFI()
ffi.cdef("""
#define SFD_CLOEXEC ...
#define SFD_NONBLOCK ...

struct signalfd_siginfo {
    uint32_t ssi_signo; /* Signal number */
    int32_t ssi_errno; /* Error number (unused) */
    int32_t ssi_code; /* Signal code */
    uint32_t ssi_pid; /* PID of sender */
    uint32_t ssi_uid; /* Real UID of sender */
    int32_t ssi_fd; /* File descriptor (SIGIO) */
    uint32_t ssi_tid; /* Kernel timer ID (POSIX timers)
    uint32_t ssi_band; /* Band event (SIGIO) */
    uint32_t ssi_overrun; /* POSIX timer overrun count */
    uint32_t ssi_trapno; /* Trap number that caused signal */
    int32_t ssi_status; /* Exit status or signal (SIGCHLD) */
    int32_t ssi_int; /* Integer sent by sigqueue(3) */
    uint64_t ssi_ptr; /* Pointer sent by sigqueue(3) */
    uint64_t ssi_utime; /* User CPU time consumed (SIGCHLD) */
    uint64_t ssi_stime; /* System CPU time consumed (SIGCHLD) */
    uint64_t ssi_addr; /* Address that generated signal
                        (for hardware-generated signals) */
//    uint8_t pad[X]; /* Pad size to 128 bytes (allow for
//                        additional fields in the future) */
    ...;
};

//#define _SIGSET_NWORDS 32
typedef struct
{
    unsigned long int __val[%d];
} __sigset_t;

typedef __sigset_t sigset_t;

int signalfd(int fd, const sigset_t *mask, int flags);

int sigemptyset(sigset_t *set);
int sigfillset(sigset_t *set);
int sigaddset(sigset_t *set, int signum);
int sigdelset(sigset_t *set, int signum);
int sigismember(const sigset_t *set, int signum);

#define SIG_BLOCK ...
#define SIG_UNBLOCK ...
#define SIG_SETMASK ...

int pthread_sigmask(int how, const sigset_t *set, sigset_t *oldset);
""" % (16 if platform.architecture()[0] == "64bit" else 32))
# define _SIGSET_NWORDS     (1024 / (8 * sizeof (unsigned long int)))
# 32bits: 1024 / 8 / 4  = 32
# 64bits: 1024 / 8 / 4  = 16

C = ffi.verify("""
#include <sys/signalfd.h>
#include <stdint.h> /* Definition of uint64_t */
#include <signal.h>
""", libraries=[], ext_package="butter")

SFD_CLOEXEC = C.SFD_CLOEXEC
SFD_NONBLOCK = C.SFD_NONBLOCK

SIG_BLOCK = C.SIG_BLOCK
SIG_UNBLOCK = C.SIG_UNBLOCK
SIG_SETMASK = C.SIG_SETMASK

NEW_SIGNALFD = -1 # Create a new signal rather than modify an exsisting one


def signalfd(signals, fd=NEW_SIGNALFD, flags=0, closefd=CLOEXEC_DEFAULT):
    """Create a new signalfd
    
    Arguments
    ----------
    :param sigset_t signals: raw cdata to pass to the syscall
    :param int signals: A single int representing the signal to listen for
    :param list signals: A list of signals to listen for
    :param int fd: The file descriptor to modify, if set to NEW_SIGNALFD then a new FD is returned
    :param int flags: 
    
    Flags
    ------
    SFD_CLOEXEC: Close the signalfd when executing a new program
    SFD_NONBLOCK: Open the socket in non-blocking mode
    
    Returns
    --------
    :return: The file descriptor representing the signalfd
    :rtype: int
    
    Exceptions
    -----------
    :raises ValueError: Invalid value in flags
    :raises OSError: Max per process FD limit reached
    :raises OSError: Max system FD limit reached
    :raises OSError: Could not mount (internal) anonymous inode device
    :raises MemoryError: Insufficient kernel memory
    """
    if hasattr(fd, 'fileno'):
        fd = fd.fileno()

    assert isinstance(fd, int), 'fd must be an integer'
    assert isinstance(flags, int), 'Flags must be an integer'

    if closefd:
        flags |= SFD_CLOEXEC

    if isinstance(signals, ffi.CData):
        mask = signals
    else:
        mask = ffi.new('sigset_t[1]')
        # if we have multiple signals then all is good
        try:
            signals = iter(signals)
        except TypeError:
        # if not make the value iterable
            signals = [signals]

        for signal in signals:
            C.sigaddset(mask, signal)

    ret_fd = C.signalfd(fd, mask, flags)
    
    if ret_fd < 0:
        err = ffi.errno
        if err == errno.EBADF:
            raise ValueError("FD is not a valid file descriptor")
        elif err == errno.EINVAL:
            if (flags & (0xffffffff ^ (SFD_CLOEXEC|SFD_NONBLOCK))):
                raise ValueError("Mask contains invalid values")
            else:
                raise ValueError("FD is not a signalfd")
        elif err == errno.EMFILE:
            raise OSError("Max system FD limit reached")
        elif err == errno.ENFILE:
            raise OSError("Max system FD limit reached")
        elif err == errno.ENODEV:
            raise OSError("Could not mount (internal) anonymous inode device")
        elif err == errno.ENOMEM:
            raise MemoryError("Insufficent kernel memory available")
        else:
            # If you are here, its a bug. send us the traceback
            raise UnknownError(err)

    return ret_fd


def pthread_sigmask(action, signals):
    """Block and Unblock signals from being delivered to an application
    
    This is required as this function/functionality does not exist in 
    python2.x
    
    Arguments
    ----------
    :param int action: The action to take on the supplied signals (bitmask)
    :param list signals: An iterable of signals
    :param int signals: A single signal
    
    Flags: action
    -----------
    SIG_BLOCK: Block the signals in sigmask from being delivered
    SIG_UNBLOCK: Unblock the signals in the supplied sigmask
    SIG_SETMASK: Set the active signals to match the supplied sigmask
    
    Returns
    --------
    :return: The old set of active signals
    :rtype: sigset
    
    Exceptions
    -----------
    :raises ValueError: Invalid value in 'action'
    :raises ValueError: sigmask is not a valid sigmask_t
    """
    assert isinstance(action, int), '"How" must be an integer'

    new_sigmask = ffi.new('sigset_t[1]')
    old_sigmask = ffi.new('sigset_t[1]')

    # if we have multiple signals then all is good
    try:
        signals = iter(signals)
    except TypeError:
    # if not make the value iterable
        signals = [signals]

    for signal in signals:
        C.sigaddset(new_sigmask, signal)
    
    ret = C.pthread_sigmask(action, new_sigmask, ffi.NULL)
    
    if ret < 0:
        err = ffi.errno
        if err == errno.EINVAL:
            raise ValueError("Action is an invalid value (not one of SIG_BLOCK, SIG_UNBLOCK or SIG_SETMASK)")
        elif err == errno.EFAULT:
            raise ValueError("sigmask is not a valid sigset_t")
        else:
            # If you are here, its a bug. send us the traceback
            raise UnknownError(err)


signum_to_signame = {val:key for key, val in signal.__dict__.items()
                     if isinstance(val, int) and "_" not in key}


#SIGINFO_LENGTH = 128 # Bytes
SIGINFO_LENGTH = ffi.sizeof('struct signalfd_siginfo')




#!/usr/bin/env python
"""system: syscalls for system managment

This module includes a number of helpful functions for managing and maintaing 
a system. These were orignially created to support a Linux contaienrs solution 
written in python but may be genrally useful and have been broken off and 
intergrarated into butter in this module


* mount: Mount filesystems using the `man 2 mount` syscall (simmilar to 
         /sbin/mount)
* umount: Unmount filesystems in the system
* pivot_root: Exchange the filesystem at 'new' with '/' and mount the old 
              filesystem at 'old'
* sethostname: Set the hostname of the system
* gethostname: Retrvie the current system hostname (identical to 
              :py:func:`socket.gethostname`)
* getpid: Call the syscall `getpid` directly, bypassing glibc and any caching 
          it performs
* getppid: Call the syscall `getppid` directly, bypassing glibc and any caching 
           it performs
"""

from __future__ import print_function

from .utils import PermissionError, InternalError, UnknownError
from os.path import isdir as _isdir
from cffi import FFI as _FFI
import errno as _errno

_ffi = _FFI()
_ffi.cdef("""
# define MS_BIND ...
# define MS_DIRSYNC ...
# define MS_MANDLOCK ...
# define MS_MOVE ...
# define MS_NOATIME ...
# define MS_NODEV ...
# define MS_NODIRATIME ...
# define MS_NOEXEC ...
# define MS_NOSUID ...
# define MS_RDONLY ...
# define MS_RELATIME ...
# define MS_REMOUNT ...
# define MS_SILENT ...
# define MS_STRICTATIME ...
# define MS_SYNCHRONOUS ...

# define MNT_FORCE ...
# define MNT_DETACH ...
# define MNT_EXPIRE ...
# define UMOUNT_NOFOLLOW ...

# define HOST_NAME_MAX ...

int mount(const char *source, const char *target,
          const char *filesystemtype, unsigned long mountflags,
          const void *data);
int umount2(const char *target, int flags);
extern int pivot_root(const char * new_root, const char * put_old);

int gethostname(char *name, size_t len);
int sethostname(const char *name, size_t len);

// Muck with the types so cffi understands it
// normmaly pid_t (defined as int32_t in
// /usr/include/arm-linux-gnueabihf/bits/typesizes.h
int32_t getpid(void);
int32_t getppid(void);
""")

_C = _ffi.verify("""  
//#include <sched.h>
#include <sys/mount.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/syscall.h>
#include <sys/mount.h>

int32_t getpid(void){
    return syscall(SYS_getpid);
};

int32_t getppid(void){
    return syscall(SYS_getppid);
};
""", libraries=[], ext_package="butter")

MS_BIND = _C.MS_BIND
MS_DIRSYNC = _C.MS_DIRSYNC
MS_MANDLOCK = _C.MS_MANDLOCK
MS_MOVE = _C.MS_MOVE
MS_NOATIME = _C.MS_NOATIME
MS_NODEV = _C.MS_NODEV
MS_NODIRATIME = _C.MS_NODIRATIME
MS_NOEXEC = _C.MS_NOEXEC
MS_NOSUID = _C.MS_NOSUID
MS_RDONLY = _C.MS_RDONLY
MS_RELATIME = _C.MS_RELATIME
MS_REMOUNT = _C.MS_REMOUNT
MS_SILENT = _C.MS_SILENT
MS_STRICTATIME = _C.MS_STRICTATIME
MS_SYNCHRONOUS = _C.MS_SYNCHRONOUS

HOST_NAME_MAX = _C.HOST_NAME_MAX

# seems reasonable
MAXPATHLEN = 256

getpid = _C.getpid
getppid = _C.getppid

class Retry(Exception):
    """Filesystem now marked as expired"""


def mount(src, target, fs, flags=0, data=""):
    """Mount the specified filesystem at `target`

    Arguments
    ----------
    :param str src: Filesystem dependent string specifing the source of the mount
                    eg for nfs this would be <ip>:/remote/path or a block device 
                    dev for a normal filesystem
    :param str target: The path to mount the filesystem on
    :param str fs: The type of filesystem to mount on `target`
    :param int flags: Extra conditions on the mount (see flags below)
    :param str data: Additinal data to pass to the filesystem driver

    Flags
    ------
    :py:const:`MS_BIND`: Make mount a bind mount, `fs` is ignored with this option
    :py:const:`MS_DIRSYNC`: Perform all directory operations synchronously
    :py:const:`MS_MANDLOCK`: Enable Mandatory locking for the filesystem
    :py:const:`MS_MOVE`: Move a mountpoint to a new location atomically without unmounting
    :py:const:`MS_NOATIME`: Dont update atime on file access
    :py:const:`MS_NODEV`: Prevent device nodes from ebing created on the filesystem
    :py:const:`MS_NODIRATIME`: Dont update Directory atime on file access
    :py:const:`MS_NOEXEC`: Prevent files from being exected on the filesystem via `exec()`
    :py:const:`MS_NOSUID`: Disable SUID flag on files in this filesystem
    :py:const:`MS_RDONLY`: Mount filesystem Read Only
    :py:const:`MS_RELATIME`: only update atime if ctime or mtime have been updated
    :py:const:`MS_REMOUNT`: Remount the filesystem in place
    :py:const:`MS_SILENT`: Disable printing messages to dmesg for the mount
    :py:const:`MS_STRICTATIME`: Always update atime on file access
    :py:const:`MS_SYNCHRONOUS`: Mount the filesystem in synchronous mode (same as passing 
                                O_SYNC to :py:func:`os.open()`

    Returns
    --------
    No return value

    Exceptions
    -----------
    :raises ValueError: Attempt to mount a Read only filesystem without specifing MS_RDONLY as a flag
    :raises ValueError: `src` or `target` contain a component that does not exist or was not searchable
    :raises ValueError: `src` is already mounted
    :raises ValueError: Filesystem cannot be mounted read only as it still holds files open for writing
    :raises ValueError: Target is busy (it is the working directory of some thread, the mount point of another device, has open files, etc.)
    :raises OSError: `src` has an invalid superblock
    :raises OSError: MS_REMOUNT was attempted but `src` is not mounted on `target`
    :raises OSError: MS_MOVE attempted but `src` is not a mount point or is '/'
    :raises OSError: Filesystem not available in the kernel
    :raises ValueError: Too many links encountered during pathname resolution
    :raises ValueError: MS_MOVE attempted while `target` is a descendent of `src`
    :raises ValueError: `src` or `target` longer than MAXPATHLEN
    :raises ValueError: `src` or `target` contains an empty or non existent component
    :raises ValueError: `src` is nto a valid block device and a block device is required by this filesystem
    :raises ValueError: `target` or prefix of `src` is nto a directory
    :raises OSError: The major number of `src` is out of the range for valid block devices
    :raises MemoryError: Kernel could not allocate enough memory to handle the request
    :raises PermissionError: No permission to mount filesystem
    """
    assert 0 < len(src) < MAXPATHLEN, "src is too long in length"
    assert 0 < len(target) < MAXPATHLEN, "target is too long in length"
    
    assert isinstance(src,    (str, bytes)), "src must be a string"
    assert isinstance(target, (str, bytes)), "target must be a string"
    assert isinstance(fs,     (str, bytes)), "fs must be a string"
    assert isinstance(flags,   int        ), "flags must be a integer"
    assert isinstance(data,   (str, bytes)), "data must be a string"

    if isinstance(src, str):
        src = src.encode()

    if isinstance(target, str):
        target = target.encode()

    if isinstance(fs, str):
        fs = fs.encode()

    if isinstance(data, str):
        data = data.encode()
    
    err = _C.mount(src, target, fs, flags, data)

    if err < 0:
        err = _ffi.errno
        if err == _errno.EACCES:
            raise PermissionError("A component of a path was not searchable. (See also path_resolution(7).) Or, mounting a read-only filesystem was attempted without giving the MS_RDONLYflag. Or, the block device source is located on a filesystem mounted with the MS_NODEV option")
        elif err == _errno.EBUSY:
            raise ValueError("source is already mounted. Or, it cannot be remounted read-only, because it still holds files open for writing. Or, it cannot be mounted on target because target is still busy (it is the working directory of some thread, the mount point of another device, has open files, etc.)")
        elif err == _errno.EFAULT:
            # In practice this should not be raised as it means this lib has passed in invalid 
            # data, this is a bug so report it if you can
            raise ValueError("One of the pointer arguments points outside the user address space")
        elif err == _errno.EINVAL:
            raise ValueError("source had an invalid superblock. Or, a remount (MS_REMOUNT) was attempted, but source was not already mounted on target. Or, a move (MS_MOVE) was attempted, but source was not a mount point, or was '/'")
        elif err == _errno.ELOOP:
            raise ValueError("Too many links encountered during pathname resolution. Or, a move was attempted, while target is a descendant of source")
        elif err == _errno.EMFILE:
            raise OSError("Table of dummy devices is full")
        elif err == _errno.ENAMETOOLONG:
            # This is checked in the assert above but check for it and report it corectly anyway
            raise ValueError("A pathname was longer than MAXPATHLEN ({})".format(MAXPATHLEN))
        elif err == _errno.ENODEV:
            raise OSError("filesystemtype not configured in the kernel")
        elif err == _errno.ENOENT:
            raise ValueError("A pathname was empty or had a nonexistent component")
        elif err == _errno.ENOMEM:
            raise MemoryError("The kernel could not allocate a free page to copy filenames or data into")
        elif err == _errno.ENOTBLK:
            raise ValueError("source is not a block device and a device was required")
        elif err == _errno.ENOTDIR:
            raise ValueError("target, or a prefix of source, is not a directory")
        elif err == _errno.ENXIO:
            raise OSError("The major number of the block device source is out of range")
        elif err == _errno.EPERM:
            raise PermissionError("Permission denied, CAP_SYS_ADMIN not in capability bits")
        else:
            # If you are here, its a bug. send us the traceback
            raise UnknownError(err)


def umount(target, flags=0):
    """Unmount the specified filesystem

    Arguments
    ----------
    :param str target: The path to the filesystem to unmount
    :param int flags: Extra options to use to unmount the filesystem

    Flags
    ------
    :py:const:`MNT_FORCE`: Forcibly detach the filesystem, even if busy (NFS only)
    :py:const:`MNT_DETACH`: Lazily detach the filesystem (filesystem will be detached
                            when there are no more consumers of the filesystem). This
                            will cause the mount to appear unmounted to processes that
                            are not using the detached mount point
    :py:const:`MNT_EXPIRE`: Mark the mountpoint as expired and trigger an EAGAIN. any
                            access by a program will mark the filesystem as active
                            again. if a filesystem is marked as expired, then another
                            umount call will unmount the filesystem normmaly
    :py:const:`UMOUNT_NOFOLLOW`: Do not derefrence any symlinks when unmounting the 
                                 filesystem

    Returns
    --------
    No return value

    Exceptions
    -----------
    :raises Retry: Filesystem now marked as expired, call again to unmount
    :raises ValueError: Could not unount filesystema s it is currently in use
    :raises OSError: Target is not a mount point
    :raises OSError: umount called with MNT_EXPIRE and ethier MNT_DETACH or MNT_FORCE
    :raises ValueError: Supplied path is too long
    :raises ValueError: Supplied path has an empty or non-existent component
    :raises MemoryError: Kernel could not allocate enough memory to handle the request
    :raises PermissionError: No permission to pivot_root to new location
    """
    assert 0 < len(target) < MAXPATHLEN, "target is too long in length"

    assert isinstance(target, (str, bytes)), "target must be a string"
    assert isinstance(flags, int), "flags must be a integer"

    if isinstance(target, str):
        target = target.encode()

    err = _C.umount2(target, flags)

    if err < 0:
        err = _ffi.errno
        if err == _errno.EAGAIN:
            raise Retry("Filesystem marked as expired, call again to unmount filesystem")
        elif err == _errno.EBUSY:
            raise ValueError("target could not be unmounted because it is busy")
        elif err == _errno.EFAULT:
            raise ValueError("target points outside the user address space")
        elif err == _errno.EINVAL:
            raise ValueError("target is not a mount point. Or, umount2() was called with MNT_EXPIRE and either MNT_DETACH or MNT_FORCE")
        elif err == _errno.ENAMETOOLONG:
            raise ValueError("A pathname was longer than MAXPATHLEN")
        elif err == _errno.ENOENT:
            raise ValueError("A pathname was empty or had a nonexistent component")
        elif err == _errno.ENOMEM:
            raise MemoryError("The kernel could not allocate a free page to copy filenames or data into")
        elif err == _errno.EPERM:
            raise PermissionError("Permission denied, CAP_SYS_ADMIN not in capability bits")
        else:
            # If you are here, its a bug. send us the traceback
            raise UnknownError(err)


def pivot_root(new, old):
    """Move the filesystem specfied by `new` and mount it at '/' and move the old '/' to `old`

    Arguments
    ----------
    :param str new: Path to a mounted filesystem to make the new '/'
    :param str old: Location where current '/' should be mounted

    Returns
    --------
    No return value

    Exceptions
    -----------
    :raises ValueError: `new` or `old` does not refer to a directory
    :raises ValueError: `new` or `old` are on the current root filesystem or filesystem already mounted on `old`
    :raises ValueError: `old` is not a folder underneath `new`
    :raises PermissionError: No permission to pivot_root to new location
    """
    assert len(new) > 0
    assert len(old) > 0

    assert isinstance(new, (str, bytes)), "new must be a string"
    assert isinstance(old, (str, bytes)), "old must be a string"

    if isinstance(new, str):
        new = new.encode()

    if isinstance(old, str):
        old = old.encode()

    err = _C.pivot_root(new, old)

    if err < 0:
        err = _ffi.errno
        if err == _errno.EINVAL:
            raise ValueError("{} is not a sub-directory of {}".format(old, new))
        elif err == _errno.EBUSY:
            raise ValueError("old or new are on the current root filesystem or filesystem already mounted on {}".format(old))
        elif err == _errno.ENOTDIR:
            if _isdir(new):
                raise ValueError("{} is not a Directory".format(new))
            elif _isdir(old):
                raise ValueError("{} is not a Directory".format(old))
            else:
                # this is a bug but testing for this case just in case, let us know if you
                # hit it
                raise ValueError("old or new is not a dir but could not work out which one")
        elif err == _errno.EPERM:
            raise PermissionError("Permission denied, CAP_SYS_ADMIN not in capability bits")
        else:
            # If you are here, its a bug. send us the traceback
            raise UnknownError(err)


def sethostname(hostname):
    """Set the hostname for they system

    Arguments
    ----------
    :param str hostname: The hostname to set

    Returns
    --------
    No return value

    Exceptions
    -----------
    :raises ValueError: Hostname too long
    :raises PermissionError: No permission to set hostname
    """
    assert len(hostname) < HOST_NAME_MAX, "Specified hostname too long"

    assert isinstance(hostname, (str, bytes)), "Hostname must be a string"

    if isinstance(hostname, str):
        hostname = hostname.encode()

    err = _C.sethostname(hostname, len(hostname))

    if err < 0:
        err = _ffi.errno
        if err == _errno.EFAULT:
            # in practice this should never be raised as it means this function is broken
            raise ValueError("Name is an invalid address")
        elif err == _errno.EINVAL:
            # same as above, we check values and supply the right ones but just in case we 
            # handle the error case
            raise ValueError("length is negative or hostname is longer than allowed value")
        elif err == _errno.EPERM:
            raise PermissionError("Permission denied, CAP_SYS_ADMIN not in capability bits")
        else:
            # If you are here, its a bug. send us the traceback
            raise UnknownError(err)

def gethostname():
    """Retrive the specified hostname of the system

    Returns
    --------
    :return: The hostname of the system
    :rtype: str
    """
    hostname = _ffi.new('char[]', HOST_NAME_MAX)
    err = _C.gethostname(hostname, len(hostname))

    if err < 0:
        err = _ffi.errno
        if err == _errno.EFAULT:
            # in practice this should never be raised as it means this function is broken
            raise ValueError("Name is an invalid address")
        elif err == _errno.EINVAL:
            # same as above, we check values and supply the right ones but just in case we 
            # handle the error case
            raise ValueError("length is negative or hostname is longer than allowed value")
        elif err == _errno.ENAMETOOLONG:
            # great, for some reason we did not allocate a long enough buffer
            # this is internal and is a bug in out code if reached
            # we allocated HOST_NAME_MAX for the length above so this should
            # be impossible to hit, using InternalError rather than ValueError as the
            # caller of this code did not provide an incorrect value but instead
            # the platform/OS provided an invalid value
            raise InternalError("Supplied buffer not long enough")
        elif err == _errno.EPERM:
            raise PermissionError("Permission denied, CAP_SYS_ADMIN not in capability bits")
        else:
            # If you are here, its a bug. send us the traceback
            raise UnknownError(err)
    
    hostname = _ffi.string(hostname, HOST_NAME_MAX)

    return hostname



#!/usr/bin/env python

from cffi import FFI
ffi = FFI()
ffi.cdef("""
int prctl(int option, unsigned long arg2, unsigned long arg3,
unsigned long arg4, unsigned long arg5);

#define PR_SET_PDEATHSIG ...
#define PR_GET_PDEATHSIG ...

#define PR_GET_DUMPABLE ...
#define PR_SET_DUMPABLE ...

#define PR_GET_UNALIGN ...
#define PR_SET_UNALIGN ...
#define PR_UNALIGN_NOPRINT ...
#define PR_UNALIGN_SIGBUS ....

#define PR_GET_KEEPCAPS ...
#define PR_SET_KEEPCAPS ...

#define PR_GET_FPEMU ...
#define PR_SET_FPEMU ...
#define PR_FPEMU_NOPRINT ...
#define PR_FPEMU_SIGFPE ...

#define PR_GET_FPEXC ...
#define PR_SET_FPEXC ...
# define PR_FP_EXC_SW_ENABLE ...
# define PR_FP_EXC_DIV ...
# define PR_FP_EXC_OVF ...
# define PR_FP_EXC_UND ...
# define PR_FP_EXC_RES ...
# define PR_FP_EXC_INV ...
# define PR_FP_EXC_DISABLED ...
# define PR_FP_EXC_NONRECOV ...
# define PR_FP_EXC_ASYNC ...
# define PR_FP_EXC_PRECISE ...

#define PR_GET_TIMING ...
#define PR_SET_TIMING ...
#define PR_TIMING_STATISTICAL ...
#define PR_TIMING_TIMESTAMP ...

#define PR_SET_NAME ...
#define PR_GET_NAME ...

#define PR_GET_ENDIAN ...
#define PR_SET_ENDIAN ...
#define PR_ENDIAN_BIG ...
#define PR_ENDIAN_LITTLE ...
#define PR_ENDIAN_PPC_LITTLE ...

#define PR_GET_SECCOMP ...
#define PR_SET_SECCOMP ...

#define PR_CAPBSET_READ ...
#define PR_CAPBSET_DROP ...

#define PR_GET_TSC ...
#define PR_SET_TSC ...
#define PR_TSC_ENABLE ...
#define PR_TSC_SIGSEGV ...

#define PR_GET_SECUREBITS ...
#define PR_SET_SECUREBITS ...

#define PR_SET_TIMERSLACK ...
#define PR_GET_TIMERSLACK ...

#define PR_TASK_PERF_EVENTS_DISABLE ...
#define PR_TASK_PERF_EVENTS_ENABLEE ...

#define PR_MCE_KILL_CLEAR ...
#define PR_MCE_KILL_SET ...

#define PR_MCE_KILL_LATE ...
#define PR_MCE_KILL_EARLY ...
#define PR_MCE_KILL_DEFAULT ...

#define PR_MCE_KILL_GET ...

#define PR_SET_MM ...
#define PR_SET_MM_START_CODE ...
#define PR_SET_MM_END_CODE ...
#define PR_SET_MM_START_DATA ...
#define PR_SET_MM_END_DATA ...
#define PR_SET_MM_START_STACK ...
#define PR_SET_MM_START_BRK ...
#define PR_SET_MM_BRK ...
#define PR_SET_MM_ARG_START ...
#define PR_SET_MM_ARG_END ...
#define PR_SET_MM_ENV_START ...
#define PR_SET_MM_ENV_END ...
#define PR_SET_MM_AUXV ...
#define PR_SET_MM_EXE_FILE ...

#define PR_SET_PTRACER ...
#define PR_SET_PTRACER_ANY ...

#define PR_SET_CHILD_SUBREAPER ...
#define PR_GET_CHILD_SUBREAPER ...

#define PR_SET_NO_NEW_PRIVS	...
#define PR_GET_NO_NEW_PRIVS	...

#define PR_GET_TID_ADDRESS ...
""")


C = ffi.verify("""  
#include <sys/prctl.h>
""", libraries=[], ext_package="butter")

### prctl

def main():
    pass

if __name__ == "__main__":
    main()



#!/usr/bih/env python
from ..inotify import Inotify as _Inotify
from collections import deque as _deque
import asyncio as _asyncio


class Inotify_async:
    def __init__(self, flags=0, *, loop=None, maxsize=0):
        self._loop = loop or _asyncio.get_event_loop()
        self._maxsize = maxsize
        
        self._inotify = inotify = _Inotify()
        
        self._getters = _deque()
        self._events = _deque()
        
    def watch(self, path, mask):
        return self._inotify.watch(path, mask)

    def ignore(self, wd):
        self._inotify.ignore(wd)
         
    @_asyncio.coroutine
    def get_event(self):
        """Remove and return an item from the queue.

        If you yield from get(), wait until a item is available.
        """
        # if event then consume it
        # otherwise register for fd callback
        #   add self to callback queue
        if self._events:
            return self._get()
        else:
            self._loop.add_reader(self._inotify.fileno(), self._read_event)

            waiter = _asyncio.Future(loop=self._loop)

            self._getters.append(waiter)

            return (yield from waiter)

    def get_event_nowait(self):
        """Remove and return an item from the queue.

        Return an item if one is immediately available, else raise QueueEmpty.
        """
        self._consume_done_putters()
        if self._events:
            return self._get()
        else:
            raise QueueEmpty

    @property
    def maxsize(self):
        """Number of items allowed in the queue."""
        return self._maxsize

    def _consume_done_getters(self):
        # Delete waiters at the head of the get() queue who've timed out.
        while self._getters and self._getters[0].done():
            self._getters.popleft()

    def _get(self):
        return self._events.popleft()

    def _put(self, item):
        self._events.append(item)

    def _read_event(self):
        """Read an event from the inotify fd and place it in the queue"""
        for event in self._inotify.read_events():
            self._put_event(event)
        
    def _put_event(self, event):
        """Put an item into the queue without blocking.

        If no free slot is immediately available, raise QueueFull.
        """
        self._consume_done_getters()
        if self._getters:
            assert not self._events, ('queue non-empty, why are getters waiting?')

            getter = self._getters.popleft()

            # Use _put and _get instead of passing item straight to getter, in
            # case a subclass has logic that must run (e.g. JoinableQueue).
            self._put(event)
            getter.set_result(self._get())
        elif self._maxsize > 0 and self._maxsize == self.qsize():
            self.loop.remove_reader(self._inotify.fileno())
            raise QueueFull
        else:
            # No listeners so queue up the data and stop 
            # listening for events on the FD
            self._put(event)
            self._loop.remove_reader(self._inotify.fileno())

    def qsize(self):
        """Returns the current size of the Queue
        
        Returns
        --------
        int: The current length of the queue
        """
        return len(self._events)
        
    def close(self):
        self._inotify.close()
    
    def __repr__(self):
        fd = self._inotify._fd or "closed"
        return "<{} fd={}>".format(self.__class__.__name__, fd)

def _watcher(loop):
    from ..inotify import IN_ALL_EVENTS
    
    inotify = Inotify_async(loop=loop)
    print(inotify)
    wd = inotify.watch('/tmp', IN_ALL_EVENTS)

    for i in range(5):
        event = yield from inotify.get_event()
        print(event)

    inotify.ignore(wd)
    print('done')

    event = yield from inotify.get_event()
    print(event)

    inotify.close()
    print(inotify)

def _main():
    import logging
    import asyncio
    
    log = logging.getLogger()
    log.setLevel(logging.DEBUG)
    log.addHandler(logging.StreamHandler())
    
    loop = asyncio.get_event_loop()
    
    from asyncio import Task
    task = Task(_watcher(loop))
    
    loop.run_until_complete(task)
    

if __name__ == "__main__":
    _main()



#!/usr/bin/env python
"""signalfd: Recive signals over a file descriptor"""

from .utils import Eventlike as _Eventlike
from .utils import CLOEXEC_DEFAULT as _CLOEXEC_DEFAULT
from ._signalfd import SFD_CLOEXEC, SFD_NONBLOCK, NEW_SIGNALFD
from ._signalfd import SIG_BLOCK, SIG_UNBLOCK, SIG_SETMASK
from ._signalfd import SIGINFO_LENGTH as _SIGINFO_LENGTH
from ._signalfd import signalfd, pthread_sigmask
from ._signalfd import signum_to_signame
from ._signalfd import ffi as _ffi, C as _C
import os as _os


class Signalfd(_Eventlike):
    def __init__(self, sigmask=set(), flags=0, closefd=_CLOEXEC_DEFAULT):
        """Create a new Signalfd object

        Arguments
        ----------
        :param int sigmask: Set of signals to respond on
        :param int flags: Flags to open the signalfd with
        
        Flags
        ------
        SFD_CLOEXEC: Close the signalfd when executing a new program
        SFD_NONBLOCK: Open the socket in non-blocking mode
        """
        super(self.__class__, self).__init__()
        
        self._flags = flags

        self._sigmask = sigmask = _ffi.new('sigset_t[1]')

        self._fd = signalfd(sigmask, NEW_SIGNALFD, flags)
        
    def __contains__(self, signal):
        val = _C.sigismember(self._sigmask, signal)
        
        if val < 0:
            raise ValueError('Signal is not a valid signal number')
        
        return bool(val)
    
    def _update(self):
        return signalfd(self._sigmask, self.fileno(), self._flags)

    def enable(self, signals):
        """Mark signals as active on signalfd
        
        :param int signals: A single signal to listen for
        :param list signals: A list of signals to listen for
        
        enable will mark all signals passed in on signalfd and takes
        ethier a signal integer representing the signal to add or an
        iterable if signals to (bulk) add
        
        The latter for is used for more efficent updates of signal masks
        """
        # if we have multipel signals then all is good
        try:
            signals = iter(signals)
        except TypeError:
        # if not make the value iterable
            signals = [signals]

        for signal in signals:
            _C.sigaddset(self._sigmask, signal)
            
        self._update()

    def enable_all(self):
        _C.sigfillset(self._sigmask)
        self._update()
        
    def disable(self, signal):
        _C.sigdelset(self._sigmask, signal)
        self._update()
        
    def disable_all(self):
        _C.sigemptyset(self._sigmask)
        self._update()
        
    def _read_events(self):
        buf = _os.read(self.fileno(), _SIGINFO_LENGTH)
        siginfo = _ffi.new('struct signalfd_siginfo *')

        _ffi.buffer(siginfo)[0:_SIGINFO_LENGTH] = buf
        
        return [Signal(siginfo)]


class Signal(object):
    """Proxy object for data returned by signalfd when a signal is recived
    
    Maps to most fields in a signalfd_siginfo
    
    Ommited mappings are ones that deal with traps and SIGQUEUE
    """
    def __init__(self, siginfo):
        self._siginfo = siginfo

    @property
    def signal(self):
        """The number of the signal being sent"""
        return self._siginfo.ssi_signo

    @property
    def signal_code(self):
        """The signal code"""
        return self._siginfo.ssi_code

    @property
    def pid(self):
        """PID of the sender of the signal"""
        return self._siginfo.ssi_pid

    @property
    def uid(self):
        """UID of the sender of the signal"""
        return self._siginfo.ssi_uid

    @property
    def sigio_fd(self):
        """FD that triggered the SIGIO signal"""
        return self._siginfo.ssi_fd

    @property
    def timer_id(self):
        """Kernel timer ID (POSIX Timers)"""
        return self._siginfo.ssi_tid
        
    @property
    def sigio_band(self):
        """SIGIO Band event"""
        return self._siginfo.ssi_band
        
    @property
    def overrun(self):
        """POSIX timer overrun count"""
        return self._siginfo.ssi_overrun
        
    @property
    def trapno(self):
        """Trap number that caused the signal"""
        return self._siginfo.trapno

    @property
    def child_status(self):
        """The return code of the exiting child process"""
        return self._siginfo.ssi_status
        
    @property
    def child_user_time(self):
        """The ammount of user time used by the exiting child process"""
        return self._siginfo.ssi_utime
        
    @property
    def child_system_time(self):
        """The ammount of system time used by the exiting child process"""
        return self._siginfo.ssi_stime

    def __repr__(self):
        # convert to Alpha name, else just return the int
        signame = signum_to_signame.get(self.signal, self.signal)
        return "<{} signal={} uid={} pid={}>".format(self.__class__.__name__, signame, self.uid, self.pid)



#!/usr/bin/env python

from butter import clone, _eventfd, _fanotify, _inotify, seccomp
from butter import _signalfd, splice, system, _timerfd, utils

import sys

name = 'butter'
path = 'butter'

## Automatically determine project version ##
from setuptools import setup, find_packages
try:
    from hgdistver import get_version
except ImportError:
    def get_version():
        import os
        
        d = {'__name__':name}

        # handle single file modules
        if os.path.isdir(path):
            module_path = os.path.join(path, '__init__.py')
        else:
            module_path = path
                                                
        with open(module_path) as f:
            try:
                exec(f.read(), None, d)
            except:
                pass

        return d.get("__version__", 0.1)

## Use py.test for "setup.py test" command ##
from setuptools.command.test import test as TestCommand
class PyTest(TestCommand):
    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True
    def run_tests(self):
        #import here, cause outside the eggs aren't loaded
        import pytest
        pytest.main(self.test_args)

## Try and extract a long description ##
readme = ""
for readme_name in ("README", "README.rst", "README.md",
                    "CHANGELOG", "CHANGELOG.rst", "CHANGELOG.md"):
    try:
        readme += open(readme_name).read() + "\n\n"
    except (OSError, IOError):
        continue

## Finally call setup ##
setup(
    name = name,
    version = get_version(),
    packages = [path, path + "/asyncio"], # corresponds to a dir 'epicworld' with a __init__.py in it
    author = "Da_Blitz",
    author_email = "code@pocketnix.org",
    maintainer=None,
    maintainer_email=None,
    description = "Library to interface to low level linux features (inotify, fanotify, timerfd, signalfd, eventfd, containers) with asyncio support",
    long_description = readme,
    license = "MIT BSD",
    keywords = "linux splice tee fanotify inotify eventfd signalfd timerfd aio clone unshare asyncio container server async",
    download_url = "http://blitz.works/butter/archive/tip.tar.bz2",
    classifiers = [
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: BSD License",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: System :: Monitoring",
        "Topic :: System :: Systems Administration",
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: System Administrators",
        "Intended Audience :: Developers",
        ],
    platforms=None,
    url = "http://blitz.works/butter/file/tip",
#    entry_points = {"console_scripts":["hammerhead=hammerhead:main",
#                                       "hammerhead-enter=hammerhead:setns"],
#                   },
#    scripts = ['scripts/dosomthing'],
    zip_safe = False,
    ext_package = name,
    ext_modules = [
        clone.ffi.verifier.get_extension(),
        _eventfd.ffi.verifier.get_extension(),
        _fanotify.ffi.verifier.get_extension(),
        _inotify.ffi.verifier.get_extension(),
        seccomp._ffi.verifier.get_extension(),
        _signalfd.ffi.verifier.get_extension(),
        splice._ffi.verifier.get_extension(),
        system._ffi.verifier.get_extension(),
        _timerfd.ffi.verifier.get_extension(),
        utils._ffi.verifier.get_extension(),
        ],
    setup_requires = [],
    install_requires = ['cffi>=0.7.2'],
    tests_require = ['tox', 'pytest', 'pytest-cov', 'pytest-mock', 'mock'],
    cmdclass = {'test': PyTest},
)



#!/usr/bih/env python
from ..signalfd import Signalfd as _Signalfd
from collections import deque as _deque
import asyncio as _asyncio

class Signalfd_async:
    def __init__(self, signals=[], flags=0, *, loop=None):
        self._loop = loop or _asyncio.get_event_loop()
        self._signalfd = _Signalfd(signals, flags)
        self._getters = _deque()
        self.enable = self._signalfd.enable
        self.enable_all = self._signalfd.enable_all
        self.disable = self._signalfd.disable
        self.disable_all = self._signalfd.disable_all
            
    @_asyncio.coroutine
    def wait(self):
        """Wait for the timer to expire, returning how many events have 
        elappsed since the last call to wait()

        Returns
        --------
        :return: The current count of the timer
        :rtype: int
        """
        self._loop.add_reader(self._signalfd.fileno(), self._read_event)

        waiter = _asyncio.Future(loop=self._loop)

        self._getters.append(waiter)

        return (yield from waiter)

    def _consume_done_getters(self):
        # Delete waiters at the head of the get() queue who've timed out.
        while self._getters and self._getters[0].done():
            self._getters.popleft()

    def _read_event(self):
        value = self._signalfd.read_event()
        self._put_event(value)
    
    def _put_event(self, value):
        """Put an item into the queue without blocking.

        If no free slot is immediately available, raise QueueFull.
        """
        self._loop.remove_reader(self._signalfd.fileno())

        self._consume_done_getters()

        for getter in self._getters:
            getter.set_result(value)

    def close(self):
        self._signalfd.close()

    def __repr__(self):
        fd = self._signalfd._fd or "closed"
        return "<{} fd={}>".format(self.__class__.__name__, fd)
 
def watcher(loop):
    from asyncio import sleep
    from ..signalfd import pthread_sigmask, SIG_BLOCK
    import signal
    import os
    
    test_signal = signal.SIGUSR1

    sfd = Signalfd_async(test_signal)
    pthread_sigmask(SIG_BLOCK, test_signal)
    sfd.enable(test_signal)
    
    print(sfd)

    count = 5
    
    for i in range(count):
        os.kill(os.getpid(), test_signal)
        sig = yield from sfd.wait()
        print(sig)

    print("Got all 5 signals")

    sfd.close()
    print(sfd)

def main():
    import logging
    import asyncio

    log = logging.getLogger()
    log.setLevel(logging.DEBUG)
    log.addHandler(logging.StreamHandler())
    
    loop = asyncio.get_event_loop()
    
    from asyncio import Task
    task = Task(watcher(loop))
    
    loop.run_until_complete(task)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
"""fanotify: wrapper around the fanotify family of syscalls for watching for file modifcation"""

from .utils import PermissionError, UnknownError, CLOEXEC_DEFAULT
from collections import namedtuple
from os import O_RDONLY, O_WRONLY, O_RDWR
from os import getpid, readlink
from os import close
from os.path import join
from cffi import FFI
import errno

READ_EVENTS_MAX = 10

ffi = FFI()
ffi.cdef("""
#define FAN_CLOEXEC ...
#define FAN_NONBLOCK ...
#define FAN_CLASS_NOTIF ...
#define FAN_CLASS_CONTENT ...
#define FAN_CLASS_PRE_CONTENT ...
#define FAN_UNLIMITED_QUEUE ...
#define FAN_UNLIMITED_MARKS ...

#define FAN_MARK_ADD ...
#define FAN_MARK_REMOVE ...
#define FAN_MARK_DONT_FOLLOW ...
#define FAN_MARK_ONLYDIR ...
#define FAN_MARK_MOUNT ...
#define FAN_MARK_IGNORED_MASK ...
#define FAN_MARK_IGNORED_SURV_MODIFY ...
#define FAN_MARK_FLUSH ...

#define FAN_ALL_MARK_FLAGS ...

#define FAN_ACCESS ...
#define FAN_MODIFY ...
#define FAN_CLOSE_WRITE ...
#define FAN_CLOSE_NOWRITE ...
#define FAN_OPEN ...
#define FAN_Q_OVERFLOW ...
#define FAN_OPEN_PERM ...
#define FAN_ACCESS_PERM ...
#define FAN_ONDIR ...
#define FAN_EVENT_ON_CHILD ...

// FAN_CLOSE_WRITE|FAN_CLOSE_NOWRITE
#define FAN_CLOSE ...

// Access control flags
#define FAN_ALLOW ...
#define FAN_DENY ...

// #define FAN_EVENT_OK ...
// #define FAN_EVENT_NEXT ...


struct fanotify_response {
    int32_t fd;
    uint32_t response;
};

//#define __aligned_u64 __u64 __attribute__((aligned(8)))
struct fanotify_event_metadata {
    uint32_t event_len;
    uint8_t vers;
    uint8_t reserved;
    uint16_t metadata_len;
    uint64_t mask;
    int32_t fd;
    int32_t pid;
};


int fanotify_init(unsigned int flags, unsigned int event_f_flags);
int fanotify_mark (int fanotify_fd, unsigned int flags, uint64_t mask, int dfd, const char *pathname);
""")

C = ffi.verify("""
#include <linux/fcntl.h>
#include <sys/fanotify.h>
""", libraries=[], ext_package="butter")

def fanotify_init(flags=0, event_flags=O_RDONLY, closefd=CLOEXEC_DEFAULT):
    """Create a fanotify handle
    """
    assert isinstance(flags, int), 'Flags must be an integer'
    assert isinstance(event_flags, int), 'Event flags must be an integer'

    if closefd:
        flags |= FAN_CLOEXEC

    fd = C.fanotify_init(flags, event_flags)

    if fd < 0:
        err = ffi.errno
        if err == errno.EINVAL:
            raise ValueError("Invalid argument or flag")
        elif err == errno.EMFILE:
            raise OSError("Maximum fanotify instances reached or cant Queue/Mark limits")
        elif err == errno.ENOMEM:
            raise MemoryError("Insufficent kernel memory avalible")
        elif err == errno.EPERM:
            raise PermissionError("Operation not permitted")
        else:
            # If you are here, its a bug. send us the traceback
            raise UnknownError(err)
                                            
    return fd

def fanotify_mark(fd, path, mask, flags, dfd=0):
    """Add a file to a fanotify context"""
    """
    EINVAL: an invalid flag or mask was passed in
    EBADF: an invalid fd was passed in
    ENOENT: directory is invalid or directory/mount not marked
    ENOMEM: no mem avalible
    ENOSPC: Too many marks
    """
    if hasattr(fd, 'fileno'):
        fd = fd.fileno()

    assert isinstance(fd, int), 'FD must be an integer'
    assert isinstance(dfd, int), 'DFD must be an integer'
    assert isinstance(flags, int), 'Flags must be an integer'
    assert isinstance(mask, int), 'Mask must be an integer'
    assert isinstance(path, (str, bytes)), 'Path must be a string'
    assert len(path) > 0, 'Path cannot be 0 chars'
    
    if isinstance(path, str):
        path = path.encode()

    ret = C.fanotify_mark(fd, flags, mask, dfd, path)
    if ret < 0:
        err = ffi.errno
        if err == errno.EINVAL:
            raise ValueError("Invalid flag or mask")
        elif err == errno.EBADF:
            raise ValueError("fd does not exist or was of the incorrect type")
        elif err == errno.ENOENT:
            raise ValueError("File pointed to by path and dfd does not exist")
        elif err == errno.ENOMEM:
            raise MemoryError("Insufficent kernel memory avalible")
        elif err == errno.ENOSPC:
            raise OSError("Too many marks")
        else:
            # If you are here, its a bug. send us the traceback
            raise UnknownError(err)

class FanotifyEvent(object):
    __slots__ = ['_filename', 'version', 'mask', 'fd', 'pid']
    def __init__(self, version, mask, fd, pid):
        self.version = version
        self.mask = mask
        self.fd = fd
        self.pid = pid

        self._filename = None
                
    @property
    def filename(self):
        if not self._filename:
            try:
                name = readlink(join('/proc', str(getpid()), 'fd', str(self.fd)))
                self._filename = name
            except OSError:
                self._filename = "<Unknown>"
    
        return self._filename
        
    def close(self):
        close(self.fd)
        self.fd = None

    def __repr__(self):
        return "<FanotifyEvent filename={}, version={}, mask=0x{:X}, fd={}, pid={}>".format(
                self.filename, self.version, self.mask, self.fd, self.pid)

    @property
    def access_event(self):
        return True if self.mask & FAN_ACCESS else False

    @property
    def access_perm_event(self):
        return True if self.mask & FAN_ACCESS_PERM else False

    @property
    def modify_event(self):
        return True if self.mask & FAN_MODIFY else False

    @property
    def close_event(self):
        return True if self.mask & FAN_CLOSE else False

    @property
    def close_write_event(self):
        return True if self.mask & FAN_CLOSE_WRITE else False

    @property
    def close_nowrite_event(self):
        return True if self.mask & FAN_CLOSE_NOWRITE else False

    @property
    def open_event(self):
        return True if self.mask & FAN_OPEN else False

    @property
    def open_perm_event(self):
        return True if self.mask & FAN_OPEN_PERM else False

    @property
    def queue_overflow_event(self):
        return True if self.mask & FAN_Q_OVERFLOW else False

    @property
    def on_dir_event(self):
        return True if self.mask & FAN_ONDIR else False

    @property
    def on_child_event(self):
        return True if self.mask & FAN_EVENT_ON_CHILD else False


def str_to_events(str):
    event_struct_size = ffi.sizeof('struct fanotify_event_metadata')

    events = []

    str_buf = ffi.new('char[]', len(str))
    str_buf[0:len(str)] = str

    i = 0
    while i < len(str_buf):
        event = ffi.cast('struct fanotify_event_metadata *', str_buf[i:i+event_struct_size])
        events.append(FanotifyEvent(event.vers, event.mask, event.fd, event.pid))

        i += event.event_len

    return events
